# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Contains a class to manage selection of a license for an Addon. """

import os
from datetime import date
from typing import Optional, Tuple

import FreeCAD
import FreeCADGui

from PySide.QtWidgets import QFileDialog, QDialog
from PySide.QtGui import QDesktopServices
from PySide.QtCore import QUrl, QFile, QIODevice

try:
    from PySide.QtGui import (
        QRegularExpressionValidator,
    )
    from PySide.QtCore import QRegularExpression

    RegexWrapper = QRegularExpression
    RegexValidatorWrapper = QRegularExpressionValidator
except ImportError:
    QRegularExpressionValidator = None
    QRegularExpression = None
    from PySide.QtGui import (
        QRegExpValidator,
    )
    from PySide.QtCore import QRegExp

    RegexWrapper = QRegExp
    RegexValidatorWrapper = QRegExpValidator

translate = FreeCAD.Qt.translate


class LicenseSelector:
    """Choose from a selection of licenses, or provide your own. Includes the capability to create
    the license file itself for a variety of popular open-source licenses, as well as providing
    links to opensource.org's page about the various licenses (which often link to other resources).
    """

    licenses = {
        "Apache-2.0": (
            "Apache License, Version 2.0",
            "https://opensource.org/licenses/Apache-2.0",
        ),
        "BSD-2-Clause": (
            "The 2-Clause BSD License",
            "https://opensource.org/licenses/BSD-2-Clause",
        ),
        "BSD-3-Clause": (
            "The 3-Clause BSD License",
            "https://opensource.org/licenses/BSD-3-Clause",
        ),
        "CC0v1": (
            "No Rights Reserved/Public Domain",
            "https://creativecommons.org/choose/zero/",
        ),
        "GPLv2": (
            "GNU General Public License version 2",
            "https://opensource.org/licenses/GPL-2.0",
        ),
        "GPLv3": (
            "GNU General Public License version 3",
            "https://opensource.org/licenses/GPL-3.0",
        ),
        "LGPLv2.1": (
            "GNU Lesser General Public License version 2.1",
            "https://opensource.org/licenses/LGPL-2.1",
        ),
        "LGPLv3": (
            "GNU Lesser General Public License version 3",
            "https://opensource.org/licenses/LGPL-3.0",
        ),
        "MIT": (
            "The MIT License",
            "https://opensource.org/licenses/MIT",
        ),
        "MPL-2.0": (
            "Mozilla Public License 2.0",
            "https://opensource.org/licenses/MPL-2.0",
        ),
    }

    def __init__(self, path_to_addon):
        self.other_label = translate(
            "AddonsInstaller",
            "Other...",
            "For providing a license other than one listed",
        )
        self.path_to_addon = path_to_addon
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "developer_mode_license.ui")
        )
        self.pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        for short_code, details in LicenseSelector.licenses.items():
            self.dialog.comboBox.addItem(
                f"{short_code}: {details[0]}", userData=short_code
            )
        self.dialog.comboBox.addItem(self.other_label)
        self.dialog.otherLineEdit.hide()
        self.dialog.otherLabel.hide()

        # Connections:
        self.dialog.comboBox.currentIndexChanged.connect(self._selection_changed)
        self.dialog.aboutButton.clicked.connect(self._about_clicked)
        self.dialog.browseButton.clicked.connect(self._browse_clicked)
        self.dialog.createButton.clicked.connect(self._create_clicked)

        # Set up the first selection to whatever the user chose last time
        short_code = self.pref.GetString("devModeLastSelectedLicense", "LGPLv2.1")
        self.set_license(short_code)

    def exec(
        self, short_code: str = None, license_path: str = ""
    ) -> Optional[Tuple[str, str]]:
        """The main method for executing this dialog, as a modal that returns a tuple of the
        license's "short code" and optionally the path to the license file. Returns a tuple
        of None,None if the user cancels the operation."""

        if short_code:
            self.set_license(short_code)
        self.dialog.pathLineEdit.setText(license_path)
        result = self.dialog.exec()
        if result == QDialog.Accepted:
            new_short_code = self.dialog.comboBox.currentData()
            new_license_path = self.dialog.pathLineEdit.text()
            if not new_short_code:
                new_short_code = self.dialog.otherLineEdit.text()
            self.pref.SetString("devModeLastSelectedLicense", new_short_code)
            return new_short_code, new_license_path
        return None

    def set_license(self, short_code):
        """Set the currently-selected license."""
        index = self.dialog.comboBox.findData(short_code)
        if index != -1:
            self.dialog.comboBox.setCurrentIndex(index)
        else:
            self.dialog.comboBox.setCurrentText(self.other_label)
            self.dialog.otherLineEdit.setText(short_code)

    def _selection_changed(self, _: int):
        """Callback: when the license selection changes, the UI is updated here."""
        if self.dialog.comboBox.currentText() == self.other_label:
            self.dialog.otherLineEdit.clear()
            self.dialog.otherLineEdit.show()
            self.dialog.otherLabel.show()
            self.dialog.aboutButton.setDisabled(True)
        else:
            self.dialog.otherLineEdit.hide()
            self.dialog.otherLabel.hide()
            self.dialog.aboutButton.setDisabled(False)

    def _current_short_code(self) -> str:
        """Gets the currently-selected license short code"""
        short_code = self.dialog.comboBox.currentData()
        if not short_code:
            short_code = self.dialog.otherLineEdit.text()
        return short_code

    def _about_clicked(self):
        """Callback: when the About button is clicked, try to launch a system-default web browser
        and display the OSI page about the currently-selected license."""
        short_code = self.dialog.comboBox.currentData()
        if short_code in LicenseSelector.licenses:
            url = LicenseSelector.licenses[short_code][1]
            QDesktopServices.openUrl(QUrl(url))
        else:
            FreeCAD.Console.PrintWarning(
                f"Internal Error: unrecognized license short code {short_code}\n"
            )

    def _browse_clicked(self):
        """Callback: browse for an existing license file."""
        start_dir = os.path.join(
            self.path_to_addon,
            self.dialog.pathLineEdit.text().replace("/", os.path.sep),
        )
        license_path, _ = QFileDialog.getOpenFileName(
            parent=self.dialog,
            caption=translate(
                "AddonsInstaller",
                "Select the corresponding license file in your Addon",
            ),
            dir=start_dir,
        )
        if license_path:
            self._set_path(self.path_to_addon, license_path)

    def _set_path(self, start_dir: str, license_path: str):
        """Sets the value displayed in the path widget to the relative path from
        start_dir to license_path"""
        license_path = license_path.replace("/", os.path.sep)
        base_dir = start_dir.replace("/", os.path.sep)
        if base_dir[-1] != os.path.sep:
            base_dir += os.path.sep
        if not license_path.startswith(base_dir):
            FreeCAD.Console.PrintError("Selected file not in Addon\n")
            # Eventually offer to copy it?
            return
        relative_path = license_path[len(base_dir) :]
        relative_path = relative_path.replace(os.path.sep, "/")
        self.dialog.pathLineEdit.setText(relative_path)

    def _create_clicked(self):
        """Asks the users for the path to save the new license file to, then copies our internal
        copy of the license text to that file."""
        start_dir = os.path.join(
            self.path_to_addon,
            self.dialog.pathLineEdit.text().replace("/", os.path.sep),
        )
        license_path, _ = QFileDialog.getSaveFileName(
            parent=self.dialog,
            caption=translate(
                "AddonsInstaller",
                "Location for new license file",
            ),
            dir=os.path.join(start_dir, "LICENSE"),
        )
        if license_path:
            self._set_path(start_dir, license_path)
            short_code = self._current_short_code()
            qf = QFile(f":/licenses/{short_code}.txt")
            if qf.exists():
                qf.open(QIODevice.ReadOnly)
                byte_data = qf.readAll()
                qf.close()

                string_data = str(byte_data, encoding="utf-8")

                if (
                    "<%%YEAR%%>" in string_data
                    or "<%%COPYRIGHT HOLDER%%>" in string_data
                ):
                    info_dlg = FreeCADGui.PySideUic.loadUi(
                        os.path.join(
                            os.path.dirname(__file__),
                            "developer_mode_copyright_info.ui",
                        )
                    )
                    info_dlg.yearLineEdit.setValidator(
                        RegexValidatorWrapper(RegexWrapper("^[12]\\d{3}$"))
                    )
                    info_dlg.yearLineEdit.setText(str(date.today().year))
                    result = info_dlg.exec()
                    if result != QDialog.Accepted:
                        return  # Don't create the file, just bail out

                    holder = info_dlg.copyrightHolderLineEdit.text()
                    year = info_dlg.yearLineEdit.text()

                    string_data = string_data.replace("<%%YEAR%%>", year)
                    string_data = string_data.replace("<%%COPYRIGHT HOLDER%%>", holder)

                with open(license_path, "w", encoding="utf-8") as f:
                    f.write(string_data)
            else:
                FreeCAD.Console.PrintError(
                    f"Cannot create license file of type {short_code}\n"
                )



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************


import unittest
import os
import shutil
import stat
import tempfile
import time
from zipfile import ZipFile
import FreeCAD

from addonmanager_git import GitManager, NoGitFound, GitFailed


class TestGit(unittest.TestCase):

    MODULE = "test_git"  # file name without extension

    def setUp(self):
        """Set up the test case: called by the unit test system"""
        self.cwd = os.getcwd()
        test_data_dir = os.path.join(
            FreeCAD.getHomePath(), "Mod", "AddonManager", "AddonManagerTest", "data"
        )
        git_repo_zip = os.path.join(test_data_dir, "test_repo.zip")
        self.test_dir = os.path.join(
            tempfile.gettempdir(), "FreeCADTesting", "AddonManagerTests", "Git"
        )
        os.makedirs(self.test_dir, exist_ok=True)
        self.test_repo_remote = os.path.join(self.test_dir, "TEST_REPO_REMOTE")
        if os.path.exists(self.test_repo_remote):
            # Make sure any old copy that got left around is deleted
            self._rmdir(self.test_repo_remote)

        if not os.path.exists(git_repo_zip):
            self.skipTest("Can't find test repo")
            return

        with ZipFile(git_repo_zip, "r") as zip_repo:
            zip_repo.extractall(self.test_repo_remote)
        self.test_repo_remote = os.path.join(self.test_repo_remote, "test_repo")

        try:
            self.git = GitManager()
        except NoGitFound:
            self.skipTest("No git found")

    def tearDown(self):
        """Clean up after the test"""
        os.chdir(self.cwd)
        # self._rmdir(self.test_dir)
        os.rename(self.test_dir, self.test_dir + ".old." + str(time.time()))

    def test_clone(self):
        """Test git clone"""
        checkout_dir = self._clone_test_repo()
        self.assertTrue(os.path.exists(checkout_dir))
        self.assertTrue(os.path.exists(os.path.join(checkout_dir, ".git")))
        self.assertEqual(
            os.getcwd(), self.cwd, "We should be left in the same CWD we started"
        )

    def test_checkout(self):
        """Test git checkout"""
        checkout_dir = self._clone_test_repo()

        self.git.checkout(checkout_dir, "HEAD~1")
        status = self.git.status(checkout_dir).strip()
        expected_status = "## HEAD (no branch)"
        self.assertEqual(status, expected_status)

        self.assertEqual(
            os.getcwd(), self.cwd, "We should be left in the same CWD we started"
        )

    def test_update(self):
        """Test using git to update the local repo"""
        checkout_dir = self._clone_test_repo()

        self.git.reset(checkout_dir, ["--hard", "HEAD~1"])
        self.assertTrue(self.git.update_available(checkout_dir))
        self.git.update(checkout_dir)
        self.assertFalse(self.git.update_available(checkout_dir))
        self.assertEqual(
            os.getcwd(), self.cwd, "We should be left in the same CWD we started"
        )

    def test_tag_and_branch(self):
        """Test checking the currently checked-out tag"""
        checkout_dir = self._clone_test_repo()

        expected_tag = "TestTag"
        self.git.checkout(checkout_dir, expected_tag)
        found_tag = self.git.current_tag(checkout_dir)
        self.assertEqual(found_tag, expected_tag)
        self.assertFalse(self.git.update_available(checkout_dir))

        expected_branch = "TestBranch"
        self.git.checkout(checkout_dir, expected_branch)
        found_branch = self.git.current_branch(checkout_dir)
        self.assertEqual(found_branch, expected_branch)
        self.assertFalse(self.git.update_available(checkout_dir))

        expected_branch = "main"
        self.git.checkout(checkout_dir, expected_branch)
        found_branch = self.git.current_branch(checkout_dir)
        self.assertEqual(found_branch, expected_branch)
        self.assertFalse(self.git.update_available(checkout_dir))

        self.assertEqual(
            os.getcwd(), self.cwd, "We should be left in the same CWD we started"
        )

    def test_get_remote(self):
        """Test getting the remote location"""
        checkout_dir = self._clone_test_repo()
        expected_remote = self.test_repo_remote
        returned_remote = self.git.get_remote(checkout_dir)
        self.assertEqual(expected_remote, returned_remote)
        self.assertEqual(
            os.getcwd(), self.cwd, "We should be left in the same CWD we started"
        )

    def test_repair(self):
        """Test the repair feature (and some exception throwing)"""
        checkout_dir = self._clone_test_repo()
        remote = self.git.get_remote(checkout_dir)
        git_dir = os.path.join(checkout_dir, ".git")
        self.assertTrue(os.path.exists(git_dir))
        self._rmdir(git_dir)

        # Make sure that we've truly broken the install
        with self.assertRaises(GitFailed):
            self.git.status(checkout_dir)

        self.git.repair(remote, checkout_dir)
        status = self.git.status(checkout_dir)
        self.assertEqual(status, "## main...origin/main\n")
        self.assertEqual(
            os.getcwd(), self.cwd, "We should be left in the same CWD we started"
        )

    def _rmdir(self, path):
        try:
            shutil.rmtree(path, onerror=self._remove_readonly)
        except Exception as e:
            print(e)

    def _remove_readonly(self, func, path, _) -> None:
        """Remove a read-only file."""

        os.chmod(path, stat.S_IWRITE)
        func(path)

    def _clone_test_repo(self) -> str:
        checkout_dir = os.path.join(self.test_dir, "test_repo")
        try:
            # Git won't clone to an existing directory, so make sure to remove it first
            if os.path.exists(checkout_dir):
                self._rmdir(checkout_dir)
            self.git.clone(self.test_repo_remote, checkout_dir)
        except GitFailed as e:
            self.fail(str(e))
        return checkout_dir



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import functools
import unittest

from PySide import QtCore, QtWidgets

import FreeCAD

from AddonManagerTest.gui.gui_mocks import (
    DialogWatcher,
    FakeWorker,
    MockThread,
)
from AddonManagerTest.app.mocks import MockAddon

from addonmanager_uninstaller_gui import AddonUninstallerGUI

translate = FreeCAD.Qt.translate


class TestUninstallerGUI(unittest.TestCase):

    MODULE = "test_uninstaller_gui"  # file name without extension

    def setUp(self):
        self.addon_to_remove = MockAddon()
        self.uninstaller_gui = AddonUninstallerGUI(self.addon_to_remove)
        self.finalized_thread = False
        self.signals_caught = []

    def tearDown(self):
        pass

    def catch_signal(self, signal_name, *_):
        self.signals_caught.append(signal_name)

    def test_confirmation_dialog_yes(self):
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Confirm remove"),
            QtWidgets.QDialogButtonBox.Yes,
        )
        answer = self.uninstaller_gui._confirm_uninstallation()
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )
        self.assertTrue(answer, "Expected a 'Yes' click to return True, but got False")

    def test_confirmation_dialog_cancel(self):
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Confirm remove"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        answer = self.uninstaller_gui._confirm_uninstallation()
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )
        self.assertFalse(
            answer, "Expected a 'Cancel' click to return False, but got True"
        )

    def test_progress_dialog(self):
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Removing Addon"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        self.uninstaller_gui._show_progress_dialog()
        # That call isn't modal, so spin our own event loop:
        while self.uninstaller_gui.progress_dialog.isVisible():
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 100)
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_timer_launches_progress_dialog(self):
        worker = FakeWorker()
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Removing Addon"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        QtCore.QTimer.singleShot(
            1000, worker.stop
        )  # If the test fails, this kills the "worker"
        self.uninstaller_gui._confirm_uninstallation = lambda: True
        self.uninstaller_gui._run_uninstaller = worker.work
        self.uninstaller_gui._finalize = lambda: None
        self.uninstaller_gui.dialog_timer.setInterval(
            1
        )  # To speed up the test, only wait 1ms
        self.uninstaller_gui.run()  # Blocks once it hits the fake worker
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )
        worker.stop()

    def test_success_dialog(self):
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Uninstall complete"),
            QtWidgets.QDialogButtonBox.Ok,
        )
        self.uninstaller_gui._succeeded(self.addon_to_remove)
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_failure_dialog(self):
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Uninstall failed"),
            QtWidgets.QDialogButtonBox.Ok,
        )
        self.uninstaller_gui._failed(
            self.addon_to_remove, "Some failure message\nAnother failure message"
        )
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_finalize(self):
        self.uninstaller_gui.finished.connect(
            functools.partial(self.catch_signal, "finished")
        )
        self.uninstaller_gui.worker_thread = MockThread()
        self.uninstaller_gui._finalize()
        self.assertIn("finished", self.signals_caught)



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Validators used for various line edits """

import keyword

from PySide.QtGui import (
    QValidator,
)

# QRegularExpressionValidator was only added at the very end of the PySide
# development cycle, so make sure to support the older QRegExp version as well.
try:
    from PySide.QtGui import (
        QRegularExpressionValidator,
    )
    from PySide.QtCore import QRegularExpression

    RegexWrapper = QRegularExpression
    RegexValidatorWrapper = QRegularExpressionValidator
except ImportError:
    QRegularExpressionValidator = None
    QRegularExpression = None
    from PySide.QtGui import (
        QRegExpValidator,
    )
    from PySide.QtCore import QRegExp

    RegexWrapper = QRegExp
    RegexValidatorWrapper = QRegExpValidator


# pylint: disable=too-few-public-methods


class NameValidator(QValidator):
    """Simple validator to exclude characters that are not valid in filenames."""

    invalid = '/\\?%*:|"<>'

    def validate(self, value: str, _: int):
        """Check the value against the validator"""
        for char in value:
            if char in NameValidator.invalid:
                return QValidator.Invalid
        return QValidator.Acceptable

    def fixup(self, value: str) -> str:
        """Remove invalid characters from value"""
        result = ""
        for char in value:
            if char not in NameValidator.invalid:
                result += char
        return result


class PythonIdentifierValidator(QValidator):
    """Validates whether input is a valid Python identifier."""

    def validate(self, value: str, _: int):
        """The function that does the validation."""
        if not value:
            return QValidator.Intermediate

        if not value.isidentifier():
            return QValidator.Invalid  # Includes an illegal character of some sort

        if keyword.iskeyword(value):
            return (
                QValidator.Intermediate
            )  # They can keep typing and it might become valid

        return QValidator.Acceptable


class SemVerValidator(RegexValidatorWrapper):
    """Implements the officially-recommended regex validator for Semantic version numbers."""

    # https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
    semver_re = RegexWrapper(
        r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
        + r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
        + r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?"
        + r"(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
    )

    def __init__(self):
        super().__init__()
        if hasattr(self, "setRegularExpression"):
            self.setRegularExpression(SemVerValidator.semver_re)
        else:
            self.setRegExp(SemVerValidator.semver_re)

    @classmethod
    def check(cls, value: str) -> bool:
        """Returns true if value validates, and false if not"""
        return cls.semver_re.match(value).hasMatch()


class CalVerValidator(RegexValidatorWrapper):
    """Implements a basic regular expression validator that makes sure an entry corresponds
    to a CalVer version numbering standard."""

    calver_re = RegexWrapper(
        r"^(?P<major>[1-9]\d{3})\.(?P<minor>[0-9]{1,2})\.(?P<patch>0|[0-9]{0,2})"
        + r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
        + r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?"
        + r"(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
    )

    def __init__(self):
        super().__init__()
        if hasattr(self, "setRegularExpression"):
            self.setRegularExpression(CalVerValidator.calver_re)
        else:
            self.setRegExp(CalVerValidator.calver_re)

    @classmethod
    def check(cls, value: str) -> bool:
        """Returns true if value validates, and false if not"""
        return cls.calver_re.match(value).hasMatch()


class VersionValidator(QValidator):
    """Implements the officially-recommended regex validator for Semantic version numbers, and a
    decent approximation of the same thing for CalVer-style version numbers."""

    def __init__(self):
        super().__init__()
        self.semver = SemVerValidator()
        self.calver = CalVerValidator()

    def validate(self, value: str, position: int):
        """Called for validation, returns a tuple of the validation state, the value, and the
        position."""
        semver_result = self.semver.validate(value, position)
        calver_result = self.calver.validate(value, position)

        if semver_result[0] == QValidator.Acceptable:
            return semver_result
        if calver_result[0] == QValidator.Acceptable:
            return calver_result
        if semver_result[0] == QValidator.Intermediate:
            return semver_result
        if calver_result[0] == QValidator.Intermediate:
            return calver_result
        return QValidator.Invalid, value, position



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2021 Russell Johnson (russ4262) <russ4262@gmail.com>    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Part
import Path.Op.Adaptive as PathAdaptive
import Path.Main.Job as PathJob
from PathTests.PathTestUtils import PathTestBase

if FreeCAD.GuiUp:
    import Path.Main.Gui.Job as PathJobGui
    import Path.Op.Gui.Adaptive as PathAdaptiveGui


class TestPathAdaptive(PathTestBase):
    """Unit tests for the Adaptive operation."""

    @classmethod
    def setUpClass(cls):
        """setUpClass()...
        This method is called upon instantiation of this test class.  Add code and objects here
        that are needed for the duration of the test() methods in this class.  In other words,
        set up the 'global' test environment here; use the `setUp()` method to set up a 'local'
        test environment.
        This method does not have access to the class `self` reference, but it
        is able to call static methods within this same class.
        """
        cls.needsInit = True


    @classmethod
    def initClass(cls):
        # Open existing FreeCAD document with test geometry
        cls.needsInit = False
        cls.doc = FreeCAD.open(
            FreeCAD.getHomePath() + "Mod/Path/PathTests/test_adaptive.fcstd"
        )

        # Create Job object, adding geometry objects from file opened above
        cls.job = PathJob.Create("Job", [cls.doc.Fusion], None)
        cls.job.GeometryTolerance.Value = 0.001
        if FreeCAD.GuiUp:
            cls.job.ViewObject.Proxy = PathJobGui.ViewProvider(cls.job.ViewObject)

        # Instantiate an Adaptive operation for querying available properties
        cls.prototype = PathAdaptive.Create("Adaptive")
        cls.prototype.Base = [(cls.doc.Fusion, ["Face3"])]
        cls.prototype.Label = "Prototype"
        _addViewProvider(cls.prototype)

        cls.doc.recompute()

    @classmethod
    def tearDownClass(cls):
        """tearDownClass()...
        This method is called prior to destruction of this test class.  Add code and objects here
        that cleanup the test environment after the test() methods in this class have been executed.
        This method does not have access to the class `self` reference.  This method
        is able to call static methods within this same class.
        """
        # FreeCAD.Console.PrintMessage("TestPathAdaptive.tearDownClass()\n")

        # Close geometry document without saving
        if not cls.needsInit:
            FreeCAD.closeDocument(cls.doc.Name)

    # Setup and tear down methods called before and after each unit test
    def setUp(self):
        """setUp()...
        This method is called prior to each `test()` method.  Add code and objects here
        that are needed for multiple `test()` methods.
        """
        if self.needsInit:
            self.initClass()

    def tearDown(self):
        """tearDown()...
        This method is called after each test() method. Add cleanup instructions here.
        Such cleanup instructions will likely undo those in the setUp() method.
        """
        pass

    # Unit tests
    def test00(self):
        """test00() Empty test."""
        return

    def test01(self):
        """test01() Verify path generated on Face3."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face3"])]  # (base, subs_list)
        adaptive.Label = "test01+"
        adaptive.Comment = "test01() Verify path generated on Face3."

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        # moves = getGcodeMoves(adaptive.Path.Commands, includeRapids=False)
        # operationMoves = ";  ".join(moves)
        # FreeCAD.Console.PrintMessage("test00_moves: " + operationMoves + "\n")

        # self.assertTrue(expected_moves_test01 == operationMoves,
        #                "expected_moves_test01: {}\noperationMoves: {}".format(expected_moves_test01, operationMoves))
        self.assertTrue(
            len(adaptive.Path.Commands) > 100, "Command count not greater than 100."
        )

    def test02(self):
        """test02() Verify path generated on adjacent, combined Face3 and Face10.  The Z heights are different."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face3", "Face10"])]  # (base, subs_list)
        adaptive.Label = "test02+"
        adaptive.Comment = "test02() Verify path generated on adjacent, combined Face3 and Face10.  The Z heights are different."

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(
            len(adaptive.Path.Commands) > 100, "Command count not greater than 100."
        )

    def test03(self):
        """test03() Verify path generated on adjacent, combined Face3 and Face10.  The Z heights are different."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face3", "Face10"])]  # (base, subs_list)
        adaptive.Label = "test03+"
        adaptive.Comment = "test03() Verify path generated on adjacent, combined Face3 and Face10.  The Z heights are different."

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = True
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(
            len(adaptive.Path.Commands) > 100, "Command count not greater than 100."
        )

    def test04(self):
        """test04() Verify path generated non-closed edges with differing Z-heights that are closed with Z=1 projection: "Edge9", "Edge2", "Edge8", "Edge15", "Edge30", "Edge31", "Edge29", "Edge19"."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [
            (
                self.doc.Fusion,
                [
                    "Edge9",
                    "Edge2",
                    "Edge8",
                    "Edge15",
                    "Edge30",
                    "Edge31",
                    "Edge29",
                    "Edge19",
                ],
            )
        ]  # (base, subs_list)
        adaptive.Label = "test04+"
        adaptive.Comment = 'test04() Verify path generated non-closed edges with differing Z-heights that are closed with Z=1 projection: "Edge9", "Edge2", "Edge8", "Edge15", "Edge30", "Edge31", "Edge29", "Edge19".'

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(
            len(adaptive.Path.Commands) > 100, "Command count not greater than 100."
        )

    def test05(self):
        """test05() Verify path generated closed wire with differing Z-heights: "Edge13", "Edge7", "Edge9", "Edge2", "Edge8", "Edge15", "Edge30", "Edge31", "Edge29", "Edge19"."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [
            (
                self.doc.Fusion,
                [
                    "Edge13",
                    "Edge7",
                    "Edge9",
                    "Edge2",
                    "Edge8",
                    "Edge15",
                    "Edge30",
                    "Edge31",
                    "Edge29",
                    "Edge19",
                ],
            )
        ]  # (base, subs_list)
        adaptive.Label = "test05+"
        adaptive.Comment = 'test05() Verify path generated closed wire with differing Z-heights: "Edge13", "Edge7", "Edge9", "Edge2", "Edge8", "Edge15", "Edge30", "Edge31", "Edge29", "Edge19".'

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(
            len(adaptive.Path.Commands) > 100, "Command count not greater than 100."
        )

    def test06(self):
        """test06() Verify path generated with outer and inner edge loops at same Z height: "Edge15", "Edge30", "Edge31", "Edge29", "Edge19", "Edge18", "Edge35", "Edge32", "Edge34", "Edge33"."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [
            (
                self.doc.Fusion,
                [
                    "Edge15",
                    "Edge30",
                    "Edge31",
                    "Edge29",
                    "Edge19",
                    "Edge18",
                    "Edge35",
                    "Edge32",
                    "Edge34",
                    "Edge33",
                ],
            )
        ]  # (base, subs_list)
        adaptive.Label = "test06+"
        adaptive.Comment = 'test06() Verify path generated with outer and inner edge loops at same Z height: "Edge15", "Edge30", "Edge31", "Edge29", "Edge19", "Edge18", "Edge35", "Edge32", "Edge34", "Edge33".'

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        # Check command count
        self.assertTrue(
            len(adaptive.Path.Commands) > 100, "Command count not greater than 100."
        )

        # Check if any paths originate inside inner hole of donut.  They should not.
        isInBox = False
        edges = [
            self.doc.Fusion.Shape.getElement(e)
            for e in ["Edge35", "Edge32", "Edge33", "Edge34"]
        ]
        square = Part.Wire(edges)
        sqrBB = square.BoundBox
        minPoint = FreeCAD.Vector(sqrBB.XMin, sqrBB.YMin, 0.0)
        maxPoint = FreeCAD.Vector(sqrBB.XMax, sqrBB.YMax, 0.0)
        for c in adaptive.Path.Commands:
            if pathOriginatesInBox(c, minPoint, maxPoint):
                isInBox = True
                break
        self.assertFalse(isInBox, "Paths originating within the inner hole.")

    def test07(self):
        """test07() Verify path generated on donut-shaped Face10."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face10"])]  # (base, subs_list)
        adaptive.Label = "test07+"
        adaptive.Comment = "test07() Verify path generated on donut-shaped Face10."

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(
            len(adaptive.Path.Commands) > 100, "Command count not greater than 100."
        )

        # Check if any paths originate inside inner hole of donut.  They should not.
        isInBox = False
        edges = [
            self.doc.Fusion.Shape.getElement(e)
            for e in ["Edge35", "Edge32", "Edge33", "Edge34"]
        ]
        square = Part.Wire(edges)
        sqrBB = square.BoundBox
        minPoint = FreeCAD.Vector(sqrBB.XMin, sqrBB.YMin, 0.0)
        maxPoint = FreeCAD.Vector(sqrBB.XMax, sqrBB.YMax, 0.0)
        for c in adaptive.Path.Commands:
            if pathOriginatesInBox(c, minPoint, maxPoint):
                isInBox = True
                break
        self.assertFalse(isInBox, "Paths originating within the inner hole.")

        # Set Adaptive op to only use the outline of the face.
        adaptive.UseOutline = True
        self.doc.recompute()

        # Check if any paths originate inside inner hole of donut.  They should not.
        isInBox = False
        edges = [
            self.doc.Fusion.Shape.getElement(e)
            for e in ["Edge35", "Edge32", "Edge33", "Edge34"]
        ]
        square = Part.Wire(edges)
        sqrBB = square.BoundBox
        minPoint = FreeCAD.Vector(sqrBB.XMin, sqrBB.YMin, 0.0)
        maxPoint = FreeCAD.Vector(sqrBB.XMax, sqrBB.YMax, 0.0)
        for c in adaptive.Path.Commands:
            if pathOriginatesInBox(c, minPoint, maxPoint):
                isInBox = True
                break
        self.assertTrue(isInBox, "No paths originating within the inner hole.")


# Eclass


def setDepthsAndHeights(op, strDep=20.0, finDep=0.0):
    """setDepthsAndHeights(op, strDep=20.0, finDep=0.0)... Sets default depths and heights for `op` passed to it"""

    # Set start and final depth in order to eliminate effects of stock (and its default values)
    op.setExpression("StartDepth", None)
    op.StartDepth.Value = strDep
    op.setExpression("FinalDepth", None)
    op.FinalDepth.Value = finDep

    # Set step down so as to only produce one layer path
    op.setExpression("StepDown", None)
    op.StepDown.Value = 20.0

    # Set Heights
    # default values used


def getGcodeMoves(cmdList, includeRapids=True, includeLines=True, includeArcs=True):
    """getGcodeMoves(cmdList, includeRapids=True, includeLines=True, includeArcs=True)...
    Accepts command dict and returns point string coordinate.
    """
    gcode_list = list()
    last = FreeCAD.Vector(0.0, 0.0, 0.0)
    for c in cmdList:
        p = c.Parameters
        name = c.Name
        if includeRapids and name in ["G0", "G00"]:
            gcode = name
            x = last.x
            y = last.y
            z = last.z
            if p.get("X"):
                x = round(p["X"], 2)
                gcode += " X" + str(x)
            if p.get("Y"):
                y = round(p["Y"], 2)
                gcode += " Y" + str(y)
            if p.get("Z"):
                z = round(p["Z"], 2)
                gcode += " Z" + str(z)
            last.x = x
            last.y = y
            last.z = z
            gcode_list.append(gcode)
        elif includeLines and name in ["G1", "G01"]:
            gcode = name
            x = last.x
            y = last.y
            z = last.z
            if p.get("X"):
                x = round(p["X"], 2)
                gcode += " X" + str(x)
            if p.get("Y"):
                y = round(p["Y"], 2)
                gcode += " Y" + str(y)
            if p.get("Z"):
                z = round(p["Z"], 2)
                gcode += " Z" + str(z)
            last.x = x
            last.y = y
            last.z = z
            gcode_list.append(gcode)
        elif includeArcs and name in ["G2", "G3", "G02", "G03"]:
            gcode = name
            x = last.x
            y = last.y
            z = last.z
            i = 0.0
            j = 0.0
            k = 0.0
            if p.get("I"):
                i = round(p["I"], 2)
            gcode += " I" + str(i)
            if p.get("J"):
                j = round(p["J"], 2)
            gcode += " J" + str(j)
            if p.get("K"):
                k = round(p["K"], 2)
            gcode += " K" + str(k)

            if p.get("X"):
                x = round(p["X"], 2)
            gcode += " X" + str(x)
            if p.get("Y"):
                y = round(p["Y"], 2)
            gcode += " Y" + str(y)
            if p.get("Z"):
                z = round(p["Z"], 2)
            gcode += " Z" + str(z)

            gcode_list.append(gcode)
            last.x = x
            last.y = y
            last.z = z
    return gcode_list


def pathOriginatesInBox(cmd, minPoint, maxPoint):
    p = cmd.Parameters
    name = cmd.Name
    if name in ["G0", "G00", "G1", "G01"]:
        if p.get("X") and p.get("Y"):
            x = p.get("X")
            y = p.get("Y")
            if x > minPoint.x and y > minPoint.y and x < maxPoint.x and y < maxPoint.y:
                return True
    return False


def _addViewProvider(adaptiveOp):
    if FreeCAD.GuiUp:
        PathOpGui = PathAdaptiveGui.PathOpGui
        cmdRes = PathAdaptiveGui.Command.res
        adaptiveOp.ViewObject.Proxy = PathOpGui.ViewProvider(
            adaptiveOp.ViewObject, cmdRes
        )


# Example string literal of expected path moves from an operation
# Expected moves for unit test01
expected_moves_test01 = "G1 X32.5 Y32.5 Z5.0;  \
G1 X17.5 Y32.5 Z5.0;  \
G1 X17.5 Y30.0 Z5.0;  \
G1 X32.5 Y30.0 Z5.0;  \
G1 X32.5 Y27.5 Z5.0;  \
G1 X17.5 Y27.5 Z5.0;  \
G1 X17.5 Y25.0 Z5.0;  \
G1 X32.5 Y25.0 Z5.0;  \
G1 X32.5 Y22.5 Z5.0;  \
G1 X17.5 Y22.5 Z5.0;  \
G1 X17.5 Y20.0 Z5.0;  \
G1 X32.5 Y20.0 Z5.0;  \
G1 X32.5 Y17.5 Z5.0;  \
G1 X17.5 Y17.5 Z5.0"



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Part
import Path
import Path.Op.Base as PathOp
import Path.Op.EngraveBase as PathEngraveBase
import PathScripts.PathUtils as PathUtils
import math
from PySide.QtCore import QT_TRANSLATE_NOOP


from PySide import QtCore

__doc__ = "Class and implementation of Path Vcarve operation"

PRIMARY = 0
SECONDARY = 1
EXTERIOR1 = 2
EXTERIOR2 = 3
COLINEAR = 4
TWIN = 5
BORDERLINE = 6

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate
_sorting = "global"


def _collectVoronoiWires(vd):
    edges = [e for e in vd.Edges if e.Color == PRIMARY]
    vertex = {}
    for e in edges:
        for v in e.Vertices:
            i = v.Index
            j = vertex.get(i, [])
            j.append(e)
            vertex[i] = j

    # knots are the start and end points of a wire
    knots = [i for i in vertex if len(vertex[i]) == 1]
    knots.extend([i for i in vertex if len(vertex[i]) > 2])
    if len(knots) == 0:
        for i in vertex:
            if len(vertex[i]) > 0:
                knots.append(i)
                break

    def consume(v, edge):
        vertex[v] = [e for e in vertex[v] if e.Index != edge.Index]
        return len(vertex[v]) == 0

    def traverse(vStart, edge, edges):
        if vStart == edge.Vertices[0].Index:
            vEnd = edge.Vertices[1].Index
            edges.append(edge)
        else:
            vEnd = edge.Vertices[0].Index
            edges.append(edge.Twin)

        consume(vStart, edge)
        if consume(vEnd, edge):
            return None
        return vEnd

    wires = []
    while knots:
        we = []
        vFirst = knots[0]
        vStart = vFirst
        vLast = vFirst
        if len(vertex[vStart]):
            while vStart is not None:
                vLast = vStart
                edges = vertex[vStart]
                if len(edges) > 0:
                    edge = edges[0]
                    vStart = traverse(vStart, edge, we)
                else:
                    vStart = None
            wires.append(we)
        if len(vertex[vFirst]) == 0:
            knots = [v for v in knots if v != vFirst]
        if len(vertex[vLast]) == 0:
            knots = [v for v in knots if v != vLast]
    return wires


def _sortVoronoiWires(wires, start=FreeCAD.Vector(0, 0, 0)):
    def closestTo(start, point):
        p = None
        l = None
        for i in point:
            if l is None or l > start.distanceToPoint(point[i]):
                l = start.distanceToPoint(point[i])
                p = i
        return (p, l)

    begin = {}
    end = {}

    for i, w in enumerate(wires):
        begin[i] = w[0].Vertices[0].toPoint()
        end[i] = w[-1].Vertices[1].toPoint()

    result = []
    while begin:
        (bIdx, bLen) = closestTo(start, begin)
        (eIdx, eLen) = closestTo(start, end)
        if bLen < eLen:
            result.append(wires[bIdx])
            start = end[bIdx]
            del begin[bIdx]
            del end[bIdx]
        else:
            result.append([e.Twin for e in reversed(wires[eIdx])])
            start = begin[eIdx]
            del begin[eIdx]
            del end[eIdx]

    return result


class _Geometry(object):
    """POD class so the limits only have to be calculated once."""

    def __init__(self, zStart, zStop, zScale):
        self.start = zStart
        self.stop = zStop
        self.scale = zScale

    @classmethod
    def FromTool(cls, tool, zStart, zFinal):
        rMax = float(tool.Diameter) / 2.0
        rMin = float(tool.TipDiameter) / 2.0
        toolangle = math.tan(math.radians(tool.CuttingEdgeAngle.Value / 2.0))
        zScale = 1.0 / toolangle
        zStop = zStart - rMax * zScale
        zOff = rMin * zScale

        return _Geometry(zStart + zOff, max(zStop + zOff, zFinal), zScale)

    @classmethod
    def FromObj(cls, obj, model):
        zStart = model.Shape.BoundBox.ZMax
        finalDepth = obj.FinalDepth.Value

        return cls.FromTool(obj.ToolController.Tool, zStart, finalDepth)


def _calculate_depth(MIC, geom):
    # given a maximum inscribed circle (MIC) and tool angle,
    # return depth of cut relative to zStart.
    depth = geom.start - round(MIC * geom.scale, 4)
    Path.Log.debug("zStart value: {} depth: {}".format(geom.start, depth))

    return max(depth, geom.stop)


def _getPartEdge(edge, depths):
    dist = edge.getDistances()
    zBegin = _calculate_depth(dist[0], depths)
    zEnd = _calculate_depth(dist[1], depths)
    return edge.toShape(zBegin, zEnd)


class ObjectVcarve(PathEngraveBase.ObjectOp):
    """Proxy class for Vcarve operation."""

    def opFeatures(self, obj):
        """opFeatures(obj) ... return all standard features and edges based geometries"""
        return (
            PathOp.FeatureTool
            | PathOp.FeatureHeights
            | PathOp.FeatureDepths
            | PathOp.FeatureBaseFaces
            | PathOp.FeatureCoolant
        )

    def setupAdditionalProperties(self, obj):
        if not hasattr(obj, "BaseShapes"):
            obj.addProperty(
                "App::PropertyLinkList",
                "BaseShapes",
                "Path",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Additional base objects to be engraved"
                ),
            )
        obj.setEditorMode("BaseShapes", 2)  # hide

    def initOperation(self, obj):
        """initOperation(obj) ... create vcarve specific properties."""
        obj.addProperty(
            "App::PropertyFloat",
            "Discretize",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property", "The deflection value for discretizing arcs"
            ),
        )
        obj.addProperty(
            "App::PropertyFloat",
            "Colinear",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Cutoff for removing colinear segments (degrees). \
                        default=10.0.",
            ),
        )
        obj.addProperty(
            "App::PropertyFloat",
            "Tolerance",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Vcarve Tolerance"),
        )
        obj.Colinear = 10.0
        obj.Discretize = 0.01
        obj.Tolerance = Path.Preferences.defaultGeometryTolerance()
        self.setupAdditionalProperties(obj)

    def opOnDocumentRestored(self, obj):
        # upgrade ...
        self.setupAdditionalProperties(obj)

    def _getPartEdges(self, obj, vWire, geom):
        edges = []
        for e in vWire:
            edges.append(_getPartEdge(e, geom))
        return edges

    def buildPathMedial(self, obj, faces):
        """constructs a medial axis path using openvoronoi"""

        def insert_many_wires(vd, wires):
            for wire in wires:
                Path.Log.debug("discretize value: {}".format(obj.Discretize))
                pts = wire.discretize(QuasiDeflection=obj.Discretize)
                ptv = [FreeCAD.Vector(p.x, p.y) for p in pts]
                ptv.append(ptv[0])

                for i in range(len(pts)):
                    vd.addSegment(ptv[i], ptv[i + 1])

        def cutWire(edges):
            path = []
            path.append(Path.Command("G0 Z{}".format(obj.SafeHeight.Value)))
            e = edges[0]
            p = e.valueAt(e.FirstParameter)
            path.append(
                Path.Command("G0 X{} Y{} Z{}".format(p.x, p.y, obj.SafeHeight.Value))
            )
            hSpeed = obj.ToolController.HorizFeed.Value
            vSpeed = obj.ToolController.VertFeed.Value
            path.append(
                Path.Command("G1 X{} Y{} Z{} F{}".format(p.x, p.y, p.z, vSpeed))
            )
            for e in edges:
                path.extend(Path.Geom.cmdsForEdge(e, hSpeed=hSpeed, vSpeed=vSpeed))

            return path

        voronoiWires = []
        for f in faces:
            vd = Path.Voronoi.Diagram()
            insert_many_wires(vd, f.Wires)

            vd.construct()

            for e in vd.Edges:
                if e.isPrimary():
                    if e.isBorderline():
                        e.Color = BORDERLINE
                    else:
                        e.Color = PRIMARY
                else:
                    e.Color = SECONDARY
            vd.colorExterior(EXTERIOR1)
            vd.colorExterior(
                EXTERIOR2,
                lambda v: not f.isInside(
                    v.toPoint(f.BoundBox.ZMin), obj.Tolerance, True
                ),
            )
            vd.colorColinear(COLINEAR, obj.Colinear)
            vd.colorTwins(TWIN)

            wires = _collectVoronoiWires(vd)
            if _sorting != "global":
                wires = _sortVoronoiWires(wires)
            voronoiWires.extend(wires)

        if _sorting == "global":
            voronoiWires = _sortVoronoiWires(voronoiWires)

        geom = _Geometry.FromObj(obj, self.model[0])

        pathlist = []
        pathlist.append(Path.Command("(starting)"))
        for w in voronoiWires:
            pWire = self._getPartEdges(obj, w, geom)
            if pWire:
                wires.append(pWire)
                pathlist.extend(cutWire(pWire))
        self.commandlist = pathlist

    def opExecute(self, obj):
        """opExecute(obj) ... process engraving operation"""
        Path.Log.track()

        if not hasattr(obj.ToolController.Tool, "CuttingEdgeAngle"):
            Path.Log.error(
                translate(
                    "Path_Vcarve",
                    "VCarve requires an engraving cutter with a cutting edge angle",
                )
            )

        if obj.ToolController.Tool.CuttingEdgeAngle >= 180.0:
            Path.Log.error(
                translate(
                    "Path_Vcarve", "Engraver cutting edge angle must be < 180 degrees."
                )
            )
            return

        try:
            faces = []

            for base in obj.BaseShapes:
                faces.extend(base.Shape.Faces)

            for base in obj.Base:
                for sub in base[1]:
                    shape = getattr(base[0].Shape, sub)
                    if isinstance(shape, Part.Face):
                        faces.append(shape)

            if not faces:
                for model in self.model:
                    if model.isDerivedFrom(
                        "Sketcher::SketchObject"
                    ) or model.isDerivedFrom("Part::Part2DObject"):
                        faces.extend(model.Shape.Faces)

            if faces:
                self.buildPathMedial(obj, faces)
            else:
                Path.Log.error(
                    translate(
                        "PathVcarve",
                        "The Job Base Object has no engraveable element. Engraving operation will produce no output.",
                    )
                )

        except Exception as e:
            Path.Log.error(
                "Error processing Base object. Engraving operation will produce no output."
            )

    def opUpdateDepths(self, obj, ignoreErrors=False):
        """updateDepths(obj) ... engraving is always done at the top most z-value"""
        job = PathUtils.findParentJob(obj)
        self.opSetDefaultValues(obj, job)

    def opSetDefaultValues(self, obj, job):
        """opSetDefaultValues(obj) ... set depths for vcarving"""
        if PathOp.FeatureDepths & self.opFeatures(obj):
            if job and len(job.Model.Group) > 0:
                bb = job.Proxy.modelBoundBox(job)
                obj.OpStartDepth = bb.ZMax
                obj.OpFinalDepth = job.Stock.Shape.BoundBox.ZMin
            else:
                obj.OpFinalDepth = -0.1

    def isToolSupported(self, obj, tool):
        """isToolSupported(obj, tool) ... returns True if v-carve op can work with tool."""
        return (
            hasattr(tool, "Diameter")
            and hasattr(tool, "CuttingEdgeAngle")
            and hasattr(tool, "TipDiameter")
        )


def SetupProperties():
    return ["Discretize"]


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns a Vcarve operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = ObjectVcarve(obj, name, parentJob)
    return obj



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2015 Dan Falck <ddfalck@gmail.com>                      *
# *   Copyright (c) 2018, 2019 Gauthier Briere                              *
# *   Copyright (c) 2019, 2020 Schildkroet                                  *
# *   Copyright (c) 2022 Larry Woestman <LarryWoestman2@gmail.com>          *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import datetime
import os

import FreeCAD
import Path.Post.Utils as PostUtils
import Path.Post.UtilsParse as PostUtilsParse
import Path.Tool.Controller as PathToolController

#
# This routine processes things in the following order:
#
#       OUTPUT_HEADER
#       SAFETYBLOCK
#       LIST_TOOLS_IN_PREAMBLE
#       PREAMBLE
#       OUTPUT_BCNC
#       SHOW_OPERATION_LABELS
#       SHOW_MACHINE_UNITS
#       PRE_OPERATION
#       ENABLE_COOLANT (coolant on)
#       operation(s)
#       POST_OPERATION
#       ENABLE_COOLANT (coolant off)
#       RETURN_TO
#       OUTPUT_BCNC
#       TOOLRETURN
#       SAFETYBLOCK
#       POSTAMBLE
#       SHOW_EDITOR
#
# The names in all caps may be enabled/disabled/modified by setting
# the corresponding value in the postprocessor.
#


def export_common(values, objectslist, filename):
    """Do the common parts of postprocessing the objects in objectslist to filename."""
    #
    nl = "\n"

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(
                f"The object {obj.Name} is not a path. Please select only path and Compounds."
            )
            return None

    # for obj in objectslist:
    #    print(obj.Name)

    print(f'PostProcessor:  {values["POSTPROCESSOR_FILE_NAME"]} postprocessing...')
    gcode = ""

    # write header
    if values["OUTPUT_HEADER"]:
        comment = PostUtilsParse.create_comment(values, "Exported by FreeCAD")
        gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
        comment = PostUtilsParse.create_comment(
            values, f'Post Processor: {values["POSTPROCESSOR_FILE_NAME"]}'
        )
        gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
        if FreeCAD.ActiveDocument:
            cam_file = os.path.basename(FreeCAD.ActiveDocument.FileName)
        else:
            cam_file = "<None>"
        comment = PostUtilsParse.create_comment(values, f"Cam File: {cam_file}")
        gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
        comment = PostUtilsParse.create_comment(
            values, f"Output Time: {str(datetime.datetime.now())}"
        )
        gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"

    # Check canned cycles for drilling
    if values["TRANSLATE_DRILL_CYCLES"]:
        if len(values["SUPPRESS_COMMANDS"]) == 0:
            values["SUPPRESS_COMMANDS"] = ["G99", "G98", "G80"]
        else:
            values["SUPPRESS_COMMANDS"] += ["G99", "G98", "G80"]

    for line in values["SAFETYBLOCK"].splitlines(False):
        gcode += f"{PostUtilsParse.linenumber(values)}{line}{nl}"

    # Write the preamble
    if values["OUTPUT_COMMENTS"]:
        if values["LIST_TOOLS_IN_PREAMBLE"]:
            for item in objectslist:
                if hasattr(item, "Proxy") and isinstance(
                    item.Proxy, PathToolController.ToolController
                ):
                    comment = PostUtilsParse.create_comment(
                        values, f"T{item.ToolNumber}={item.Name}"
                    )
                    gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
        comment = PostUtilsParse.create_comment(values, "Begin preamble")
        gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
    for line in values["PREAMBLE"].splitlines(False):
        gcode += f"{PostUtilsParse.linenumber(values)}{line}{nl}"
    # verify if PREAMBLE or SAFETYBLOCK have changed MOTION_MODE or UNITS
    if "G90" in values["PREAMBLE"] or "G90" in values["SAFETYBLOCK"]:
        values["MOTION_MODE"] = "G90"
    elif "G91" in values["PREAMBLE"] or "G91" in values["SAFETYBLOCK"]:
        values["MOTION_MODE"] = "G91"
    else:
        gcode += f'{PostUtilsParse.linenumber(values)}{values["MOTION_MODE"]}{nl}'
    if "G21" in values["PREAMBLE"] or "G21" in values["SAFETYBLOCK"]:
        values["UNITS"] = "G21"
        values["UNIT_FORMAT"] = "mm"
        values["UNIT_SPEED_FORMAT"] = "mm/min"
    elif "G20" in values["PREAMBLE"] or "G20" in values["SAFETYBLOCK"]:
        values["UNITS"] = "G20"
        values["UNIT_FORMAT"] = "in"
        values["UNIT_SPEED_FORMAT"] = "in/min"
    else:
        gcode += f'{PostUtilsParse.linenumber(values)}{values["UNITS"]}{nl}'

    for obj in objectslist:

        # Skip inactive operations
        if hasattr(obj, "Active") and not obj.Active:
            continue
        if hasattr(obj, "Base") and hasattr(obj.Base, "Active") and not obj.Base.Active:
            continue

        # do the pre_op
        if values["OUTPUT_BCNC"]:
            comment = PostUtilsParse.create_comment(values, f"Block-name: {obj.Label}")
            gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
            comment = PostUtilsParse.create_comment(values, "Block-expand: 0")
            gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
            comment = PostUtilsParse.create_comment(values, "Block-enable: 1")
            gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
        if values["OUTPUT_COMMENTS"]:
            if values["SHOW_OPERATION_LABELS"]:
                comment = PostUtilsParse.create_comment(
                    values, f"Begin operation: {obj.Label}"
                )
            else:
                comment = PostUtilsParse.create_comment(values, "Begin operation")
            gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
            if values["SHOW_MACHINE_UNITS"]:
                comment = PostUtilsParse.create_comment(
                    values, f'Machine units: {values["UNIT_SPEED_FORMAT"]}'
                )
                gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
            if values["OUTPUT_MACHINE_NAME"]:
                comment = PostUtilsParse.create_comment(
                    values,
                    f'Machine: {values["MACHINE_NAME"]}, {values["UNIT_SPEED_FORMAT"]}',
                )
                gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
        for line in values["PRE_OPERATION"].splitlines(False):
            gcode += f"{PostUtilsParse.linenumber(values)}{line}{nl}"

        # get coolant mode
        coolantMode = "None"
        if (
            hasattr(obj, "CoolantMode")
            or hasattr(obj, "Base")
            and hasattr(obj.Base, "CoolantMode")
        ):
            if hasattr(obj, "CoolantMode"):
                coolantMode = obj.CoolantMode
            else:
                coolantMode = obj.Base.CoolantMode

        # turn coolant on if required
        if values["ENABLE_COOLANT"]:
            if values["OUTPUT_COMMENTS"] and coolantMode != "None":
                comment = PostUtilsParse.create_comment(
                    values, f"Coolant On: {coolantMode}"
                )
                gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
            if coolantMode == "Flood":
                gcode += f"{PostUtilsParse.linenumber(values)}M8{nl}"
            elif coolantMode == "Mist":
                gcode += f"{PostUtilsParse.linenumber(values)}M7{nl}"

        # process the operation gcode
        gcode += PostUtilsParse.parse_a_group(values, obj)
        # do the post_op
        if values["OUTPUT_COMMENTS"]:
            comment = PostUtilsParse.create_comment(
                values, f'{values["FINISH_LABEL"]} operation: {obj.Label}'
            )
            gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
        for line in values["POST_OPERATION"].splitlines(False):
            gcode += f"{PostUtilsParse.linenumber(values)}{line}{nl}"

        # turn coolant off if required
        if values["ENABLE_COOLANT"] and coolantMode != "None":
            if values["OUTPUT_COMMENTS"]:
                comment = PostUtilsParse.create_comment(
                    values, f"Coolant Off: {coolantMode}"
                )
                gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
            gcode += f"{PostUtilsParse.linenumber(values)}M9{nl}"

    if values["RETURN_TO"]:
        num_x = values["RETURN_TO"][0]
        num_y = values["RETURN_TO"][1]
        num_z = values["RETURN_TO"][2]
        gcode += f"{PostUtilsParse.linenumber(values)}G0 X{num_x} Y{num_y} Z{num_z}{nl}"

    # do the post_amble
    if values["OUTPUT_BCNC"]:
        comment = PostUtilsParse.create_comment(values, "Block-name: post_amble")
        gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
        comment = PostUtilsParse.create_comment(values, "Block-expand: 0")
        gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
        comment = PostUtilsParse.create_comment(values, "Block-enable: 1")
        gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
    if values["OUTPUT_COMMENTS"]:
        comment = PostUtilsParse.create_comment(values, "Begin postamble")
        gcode += f"{PostUtilsParse.linenumber(values)}{comment}{nl}"
    for line in values["TOOLRETURN"].splitlines(False):
        gcode += f"{PostUtilsParse.linenumber(values)}{line}{nl}"
    for line in values["SAFETYBLOCK"].splitlines(False):
        gcode += f"{PostUtilsParse.linenumber(values)}{line}{nl}"
    for line in values["POSTAMBLE"].splitlines(False):
        gcode += f"{PostUtilsParse.linenumber(values)}{line}{nl}"

    if FreeCAD.GuiUp and values["SHOW_EDITOR"]:
        final = gcode
        if len(gcode) > 100000:
            print("Skipping editor since output is greater than 100kb")
        else:
            dia = PostUtils.GCodeEditorDialog()
            dia.editor.setText(gcode)
            result = dia.exec_()
            if result:
                final = dia.editor.toPlainText()
    else:
        final = gcode

    print("done postprocessing.")

    if not filename == "-":
        with open(filename, "w", newline=values["END_OF_LINE_CHARACTERS"]) as gfile:
            gfile.write(final)

    return final



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Class to guess metadata based on folder contents. Note that one of the functions
of this file is to guess the license being applied to the new software package based
in its contents. It is up to the user to make the final determination about whether
the selected license is the correct one, and inclusion here shouldn't be construed as
endorsement of any particular license. In addition, the inclusion of those text strings
does not imply a modification to the license for THIS software, which is licensed
under the LGPLv2.1 license (as stated above)."""

import datetime
import os

import FreeCAD
from addonmanager_git import initialize_git, GitManager
from addonmanager_utilities import get_readme_url

translate = FreeCAD.Qt.translate

# pylint: disable=too-few-public-methods


class AddonSlice:
    """A tiny class to implement duck-typing for the URL-parsing utility functions"""

    def __init__(self, url, branch):
        self.url = url
        self.branch = branch


class Predictor:
    """Guess the appropriate metadata to apply to a project based on various parameters
    found in the supplied directory."""

    def __init__(self):
        self.path = None
        self.metadata = FreeCAD.Metadata()
        self.license_data = None
        self.readme_data = None
        self.license_file = ""
        self.git_manager: GitManager = initialize_git()
        if not self.git_manager:
            raise Exception("Cannot use Developer Mode without git installed")

    def predict_metadata(self, path: str) -> FreeCAD.Metadata:
        """Create a predicted Metadata object based on the contents of the passed-in directory"""
        if not os.path.isdir(path):
            return None
        self.path = path
        self._predict_author_info()
        self._predict_name()
        self._predict_description()
        self._predict_contents()
        self._predict_icon()
        self._predict_urls()
        self._predict_license()
        self._predict_version()

        return self.metadata

    def _predict_author_info(self):
        """Look at the git commit history and attempt to discern maintainer and author
        information."""

        committers = self.git_manager.get_last_committers(self.path)

        # This is a dictionary keyed to the author's name (which can be many
        # things, depending on the author) containing two fields, "email" and "count". It
        # is common for there to be multiple entries representing the same human being,
        # so a passing attempt is made to reconcile:
        filtered_committers = {}
        for key, committer in committers.items():
            if "github" in key.lower():
                # Robotic merge commit (or other similar), ignore
                continue
            # Does any other committer share any of these emails?
            for other_key, other_committer in committers.items():
                if other_key == key:
                    continue
                for other_email in other_committer["email"]:
                    if other_email in committer["email"]:
                        # There is overlap in the two email lists, so this is probably the
                        # same author, with a different name (username, pseudonym, etc.)
                        if not committer["aka"]:
                            committer["aka"] = set()
                        committer["aka"].add(other_key)
                        committer["count"] += other_committer["count"]
                        committer["email"].combine(other_committer["email"])
                        committers.pop(other_key)
                        break
            filtered_committers[key] = committer
        maintainers = []
        for name, info in filtered_committers.items():
            if "aka" in info:
                for other_name in info["aka"]:
                    # Heuristic: the longer name is more likely to be the actual legal name
                    if len(other_name) > len(name):
                        name = other_name
            # There is no logical basis to choose one email address over another, so just
            # take the first one
            email = info["email"][0]
            commit_count = info["count"]
            maintainers.append({"name": name, "email": email, "count": commit_count})

        # Sort by count of commits
        maintainers.sort(key=lambda i: i["count"], reverse=True)

        self.metadata.Maintainer = maintainers

    def _predict_name(self):
        """Predict the name based on the local path name and/or the contents of a
        README.md file."""

        normed_path = self.path.replace(os.path.sep, "/")
        path_components = normed_path.split("/")
        final_path_component = path_components[-1]
        predicted_name = final_path_component.replace("/", "")
        self.metadata.Name = predicted_name

    def _predict_description(self):
        """Predict the description based on the contents of a README.md file."""
        self._load_readme()

        if not self.readme_data:
            return

        lines = self.readme_data.split("\n")
        description = ""
        for line in lines:
            if "#" in line:
                continue  # Probably not a line of description
            if "![" in line:
                continue  # An image link, probably separate from any description
            if not line and description:
                break  # We're done: this is a blank line, and we've read some data already
            if description:
                description += " "
            description += line

        if description:
            self.metadata.Description = description

    def _predict_contents(self):
        """Predict the contents based on the contents of the directory."""

    def _predict_icon(self):
        """Predict the icon based on either a class which defines an Icon member, or
        the contents of the local directory structure."""

    def _predict_urls(self):
        """Predict the URLs based on git settings"""

        branch = self.git_manager.current_branch(self.path)
        remote = self.git_manager.get_remote(self.path)

        addon = AddonSlice(remote, branch)
        readme = get_readme_url(addon)

        self.metadata.addUrl("repository", remote, branch)
        self.metadata.addUrl("readme", readme)

    def _predict_license(self):
        """Predict the license based on any existing license file."""

        # These are processed in order, so the BSD 3 clause must come before the 2, for example,
        # because the only difference between them is the additional clause.
        known_strings = {
            "Apache-2.0": (
                "Apache License, Version 2.0",
                "Apache License\nVersion 2.0, January 2004",
            ),
            "BSD-3-Clause": (
                "The 3-Clause BSD License",
                "3. Neither the name of the copyright holder nor the names of its contributors \
                may be used to endorse or promote products derived from this software without \
                specific prior written permission.",
            ),
            "BSD-2-Clause": (
                "The 2-Clause BSD License",
                "2. Redistributions in binary form must reproduce the above copyright notice, \
                this list of conditions and the following disclaimer in the documentation and/or \
                other materials provided with the distribution.",
            ),
            "CC0v1": (
                "CC0 1.0 Universal",
                "voluntarily elects to apply CC0 to the Work and publicly distribute the Work \
                under its terms",
            ),
            "GPLv2": (
                "GNU General Public License version 2",
                "GNU GENERAL PUBLIC LICENSE\nVersion 2, June 1991",
            ),
            "GPLv3": (
                "GNU General Public License version 3",
                "The GNU General Public License is a free, copyleft license for software and \
                other kinds of works.",
            ),
            "LGPLv2.1": (
                "GNU Lesser General Public License version 2.1",
                "GNU Lesser General Public License\nVersion 2.1, February 1999",
            ),
            "LGPLv3": (
                "GNU Lesser General Public License version 3",
                "GNU LESSER GENERAL PUBLIC LICENSE\nVersion 3, 29 June 2007",
            ),
            "MIT": (
                "The MIT License",
                "including without limitation the rights to use, copy, modify, merge, publish, \
                distribute, sublicense, and/or sell copies of the Software",
            ),
            "MPL-2.0": (
                "Mozilla Public License 2.0",
                "https://opensource.org/licenses/MPL-2.0",
            ),
        }
        self._load_license()
        if self.license_data:
            for shortcode, test_data in known_strings.items():
                if shortcode.lower() in self.license_data.lower():
                    self.metadata.addLicense(shortcode, self.license_file)
                    return
                for test_text in test_data:
                    # Do the comparison without regard to whitespace or capitalization
                    if (
                        "".join(test_text.split()).lower()
                        in "".join(self.license_data.split()).lower()
                    ):
                        self.metadata.addLicense(shortcode, self.license_file)
                        return

    def _predict_version(self):
        """Default to a CalVer style set to today's date"""
        year = datetime.date.today().year
        month = datetime.date.today().month
        day = datetime.date.today().day
        version_string = f"{year}.{month:>02}.{day:>02}"
        self.metadata.Version = version_string

    def _load_readme(self):
        """Load in any existing readme"""
        valid_names = ["README.md", "README.txt", "README"]
        for name in valid_names:
            full_path = os.path.join(self.path, name)
            if os.path.exists(full_path):
                with open(full_path, encoding="utf-8") as f:
                    self.readme_data = f.read()
                    return

    def _load_license(self):
        """Load in any existing license"""
        valid_names = [
            "LICENSE",
            "LICENCE",
            "COPYING",
            "LICENSE.txt",
            "LICENCE.txt",
            "COPYING.txt",
        ]
        for name in valid_names:
            full_path = os.path.join(self.path.replace("/", os.path.sep), name)
            if os.path.isfile(full_path):
                with open(full_path, encoding="utf-8") as f:
                    self.license_data = f.read()
                    self.license_file = name
                    return



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Class to manage the display of an Update All dialog."""

from enum import IntEnum, auto
import os
from typing import List

import FreeCAD
import FreeCADGui

from PySide import QtCore, QtWidgets

from Addon import Addon

from addonmanager_installer import AddonInstaller, MacroInstaller

translate = FreeCAD.Qt.translate

# pylint: disable=too-few-public-methods,too-many-instance-attributes


class UpdaterFactory:
    """A factory class for generating updaters. Mainly exists to allow easily mocking
    those updaters during testing. A replacement class need only provide a
    "get_updater" function that returns mock updater objects. Those objects must be
    QObjects with a run() function and a finished signal."""

    def __init__(self, addons):
        self.addons = addons

    def get_updater(self, addon):
        """Get an updater for this addon (either a MacroInstaller or an
        AddonInstaller)"""
        if addon.macro is not None:
            return MacroInstaller(addon)
        return AddonInstaller(addon, self.addons)


class AddonStatus(IntEnum):
    """The current status of the installation process for a given addon"""

    WAITING = auto()
    INSTALLING = auto()
    SUCCEEDED = auto()
    FAILED = auto()

    def ui_string(self):
        """Get the string that the UI should show for this status"""
        if self.value == AddonStatus.WAITING:
            return ""
        if self.value == AddonStatus.INSTALLING:
            return translate("AddonsInstaller", "Installing") + "..."
        if self.value == AddonStatus.SUCCEEDED:
            return translate("AddonsInstaller", "Succeeded")
        if self.value == AddonStatus.FAILED:
            return translate("AddonsInstaller", "Failed")
        return "[INTERNAL ERROR]"


class UpdateAllGUI(QtCore.QObject):
    """A GUI to display and manage an "update all" process."""

    finished = QtCore.Signal()
    addon_updated = QtCore.Signal(object)

    def __init__(self, addons: List[Addon]):
        super().__init__()
        self.addons = addons
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "update_all.ui")
        )
        self.dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)
        self.row_map = {}
        self.in_process_row = None
        self.active_installer = None
        self.addons_with_update: List[Addon] = []
        self.updater_factory = UpdaterFactory(addons)
        self.worker_thread = None
        self.running = False
        self.cancelled = False

    def run(self):
        """Run the Update All process. Blocks until updates are complete or
        cancelled."""
        self.running = True
        self._setup_dialog()
        self.dialog.show()
        self._process_next_update()

    def _setup_dialog(self):
        """Prepare the dialog for display"""
        self.dialog.rejected.connect(self._cancel_installation)
        self.dialog.tableWidget.clear()
        self.in_process_row = None
        self.row_map = {}
        for addon in self.addons:
            if addon.status() == Addon.Status.UPDATE_AVAILABLE:
                self._add_addon_to_table(addon)
                self.addons_with_update.append(addon)

    def _cancel_installation(self):
        self.cancelled = True
        if self.worker_thread and self.worker_thread.isRunning():
            self.worker_thread.requestInterruption()

    def _add_addon_to_table(self, addon: Addon):
        """Add the given addon to the list, with no icon in the first column"""
        new_row = self.dialog.tableWidget.rowCount()
        self.dialog.tableWidget.setColumnCount(2)
        self.dialog.tableWidget.setRowCount(new_row + 1)
        self.dialog.tableWidget.setItem(
            new_row, 0, QtWidgets.QTableWidgetItem(addon.display_name)
        )
        self.dialog.tableWidget.setItem(new_row, 1, QtWidgets.QTableWidgetItem(""))
        self.row_map[addon.name] = new_row

    def _update_addon_status(self, row: int, status: AddonStatus):
        """Update the GUI to reflect this addon's new status."""
        self.dialog.tableWidget.item(row, 1).setText(status.ui_string())

    def _process_next_update(self):
        """Grab the next addon in the list and start its updater."""
        if self.addons_with_update:
            addon = self.addons_with_update.pop(0)
            self.in_process_row = (
                self.row_map[addon.name] if addon.name in self.row_map else None
            )
            self._update_addon_status(self.in_process_row, AddonStatus.INSTALLING)
            self.dialog.tableWidget.scrollToItem(
                self.dialog.tableWidget.item(self.in_process_row, 0)
            )
            self.active_installer = self.updater_factory.get_updater(addon)
            self._launch_active_installer()
        else:
            self._finalize()

    def _launch_active_installer(self):
        """Set up and run the active installer in a new thread."""

        self.active_installer.success.connect(self._update_succeeded)
        self.active_installer.failure.connect(self._update_failed)
        self.active_installer.finished.connect(self._update_finished)

        self.worker_thread = QtCore.QThread()
        self.active_installer.moveToThread(self.worker_thread)
        self.worker_thread.started.connect(self.active_installer.run)
        self.worker_thread.start()

    def _update_succeeded(self, addon):
        """Callback for a successful update"""
        self._update_addon_status(self.row_map[addon.name], AddonStatus.SUCCEEDED)
        self.addon_updated.emit(addon)

    def _update_failed(self, addon):
        """Callback for a failed update"""
        self._update_addon_status(self.row_map[addon.name], AddonStatus.FAILED)

    def _update_finished(self):
        """Callback for updater that has finished all its work"""
        if self.worker_thread is not None and self.worker_thread.isRunning():
            self.worker_thread.quit()
            self.worker_thread.wait()
        self.addon_updated.emit(self.active_installer.addon_to_install)
        if not self.cancelled:
            self._process_next_update()
        else:
            self._setup_cancelled_state()

    def _finalize(self):
        """No more updates, clean up and shut down"""
        if self.worker_thread is not None and self.worker_thread.isRunning():
            self.worker_thread.quit()
            self.worker_thread.wait()
        text = translate("Addons installer", "Finished updating the following addons")
        self._set_dialog_to_final_state(text)
        self.running = False
        self.finished.emit()

    def _setup_cancelled_state(self):
        text1 = translate("AddonsInstaller", "Update was cancelled")
        text2 = translate("AddonsInstaller", "some addons may have been updated")
        self._set_dialog_to_final_state(text1 + ": " + text2)
        self.running = False
        self.finished.emit()

    def _set_dialog_to_final_state(self, new_content):
        self.dialog.buttonBox.clear()
        self.dialog.buttonBox.addButton(QtWidgets.QDialogButtonBox.Close)
        self.dialog.label.setText(new_content)

    def is_running(self):
        """True if the thread is running, and False if not"""
        return self.running



# SPDX-License-Identifier: LGPL-2.1-or-later
# FreeCAD init script of the AddonManager module
# (c) 2001 Juergen Riegel
# License LGPL

import FreeCAD

FreeCAD.__unit_test__ += ["TestAddonManagerApp"]



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Contains a the Addon Manager's preferences dialog management class """

import os

import FreeCAD
import FreeCADGui

from PySide import QtCore
from PySide.QtGui import QIcon
from PySide.QtWidgets import (
    QWidget,
    QCheckBox,
    QComboBox,
    QDialog,
    QHeaderView,
    QRadioButton,
    QLineEdit,
    QTextEdit,
)

translate = FreeCAD.Qt.translate

# pylint: disable=too-few-public-methods


class AddonManagerOptions:
    """A class containing a form element that is inserted as a FreeCAD preference page."""

    def __init__(self, _=None):
        self.form = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "AddonManagerOptions.ui")
        )
        self.table_model = CustomRepoDataModel()
        self.form.customRepositoriesTableView.setModel(self.table_model)

        self.form.addCustomRepositoryButton.setIcon(
            QIcon.fromTheme("add", QIcon(":/icons/list-add.svg"))
        )
        self.form.removeCustomRepositoryButton.setIcon(
            QIcon.fromTheme("remove", QIcon(":/icons/list-remove.svg"))
        )

        self.form.customRepositoriesTableView.horizontalHeader().setStretchLastSection(
            False
        )
        self.form.customRepositoriesTableView.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.Stretch
        )
        self.form.customRepositoriesTableView.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.ResizeToContents
        )

        self.form.addCustomRepositoryButton.clicked.connect(
            self._add_custom_repo_clicked
        )
        self.form.removeCustomRepositoryButton.clicked.connect(
            self._remove_custom_repo_clicked
        )
        self.form.customRepositoriesTableView.doubleClicked.connect(
            self._row_double_clicked
        )

    def saveSettings(self):
        """Required function: called by the preferences dialog when Apply or Save is clicked,
        saves out the preference data by reading it from the widgets."""
        for widget in self.form.children():
            self.recursive_widget_saver(widget)
        self.table_model.save_model()

    def recursive_widget_saver(self, widget):
        """Writes out the data for this widget and all of its children, recursively."""
        if isinstance(widget, QWidget):
            # See if it's one of ours:
            pref_path = widget.property("prefPath")
            pref_entry = widget.property("prefEntry")
            if pref_path and pref_entry:
                pref_path = pref_path.data()
                pref_entry = pref_entry.data()
                pref_access_string = (
                    f"User parameter:BaseApp/Preferences/{str(pref_path,'utf-8')}"
                )
                pref = FreeCAD.ParamGet(pref_access_string)
                if isinstance(widget, QCheckBox):
                    checked = widget.isChecked()
                    pref.SetBool(str(pref_entry, "utf-8"), checked)
                elif isinstance(widget, QRadioButton):
                    checked = widget.isChecked()
                    pref.SetBool(str(pref_entry, "utf-8"), checked)
                elif isinstance(widget, QComboBox):
                    new_index = widget.currentIndex()
                    pref.SetInt(str(pref_entry, "utf-8"), new_index)
                elif isinstance(widget, QTextEdit):
                    text = widget.toPlainText()
                    pref.SetString(str(pref_entry, "utf-8"), text)
                elif isinstance(widget, QLineEdit):
                    text = widget.text()
                    pref.SetString(str(pref_entry, "utf-8"), text)
                elif widget.metaObject().className() == "Gui::PrefFileChooser":
                    filename = str(widget.property("fileName"))
                    filename = pref.SetString(str(pref_entry, "utf-8"), filename)

        # Recurse over children
        if isinstance(widget, QtCore.QObject):
            for child in widget.children():
                self.recursive_widget_saver(child)

    def loadSettings(self):
        """Required function: called by the preferences dialog when it is launched,
        loads the preference data and assigns it to the widgets."""
        for widget in self.form.children():
            self.recursive_widget_loader(widget)
        self.table_model.load_model()

    def recursive_widget_loader(self, widget):
        """Loads the data for this widget and all of its children, recursively."""
        if isinstance(widget, QWidget):
            # See if it's one of ours:
            pref_path = widget.property("prefPath")
            pref_entry = widget.property("prefEntry")
            if pref_path and pref_entry:
                pref_path = pref_path.data()
                pref_entry = pref_entry.data()
                pref_access_string = (
                    f"User parameter:BaseApp/Preferences/{str(pref_path,'utf-8')}"
                )
                pref = FreeCAD.ParamGet(pref_access_string)
                if isinstance(widget, QCheckBox):
                    widget.setChecked(pref.GetBool(str(pref_entry, "utf-8")))
                elif isinstance(widget, QRadioButton):
                    if pref.GetBool(str(pref_entry, "utf-8")):
                        widget.setChecked(True)
                elif isinstance(widget, QComboBox):
                    new_index = pref.GetInt(str(pref_entry, "utf-8"))
                    widget.setCurrentIndex(new_index)
                elif isinstance(widget, QTextEdit):
                    text = pref.GetString(str(pref_entry, "utf-8"))
                    widget.setText(text)
                elif isinstance(widget, QLineEdit):
                    text = pref.GetString(str(pref_entry, "utf-8"))
                    widget.setText(text)
                elif widget.metaObject().className() == "Gui::PrefFileChooser":
                    filename = pref.GetString(str(pref_entry, "utf-8"))
                    widget.setProperty("fileName", filename)

        # Recurse over children
        if isinstance(widget, QtCore.QObject):
            for child in widget.children():
                self.recursive_widget_loader(child)

    def _add_custom_repo_clicked(self):
        """Callback: show the Add custom repo dialog"""
        dlg = CustomRepositoryDialog()
        url, branch = dlg.exec()
        if url and branch:
            self.table_model.appendData(url, branch)

    def _remove_custom_repo_clicked(self):
        """Callback: when the remove button is clicked, get the current selection and remove it."""
        item = self.form.customRepositoriesTableView.currentIndex()
        if not item.isValid():
            return
        row = item.row()
        self.table_model.removeRows(row, 1, QtCore.QModelIndex())

    def _row_double_clicked(self, item):
        """Edit the row that was double-clicked"""
        row = item.row()
        dlg = CustomRepositoryDialog()
        url_index = self.table_model.createIndex(row, 0)
        branch_index = self.table_model.createIndex(row, 1)
        dlg.dialog.urlLineEdit.setText(self.table_model.data(url_index))
        dlg.dialog.branchLineEdit.setText(self.table_model.data(branch_index))
        url, branch = dlg.exec()
        if url and branch:
            self.table_model.setData(url_index, url)
            self.table_model.setData(branch_index, branch)


class CustomRepoDataModel(QtCore.QAbstractTableModel):
    """The model for the custom repositories: wraps the underlying preference data and uses that
    as its main data store."""

    def __init__(self):
        super().__init__()
        pref_access_string = "User parameter:BaseApp/Preferences/Addons"
        self.pref = FreeCAD.ParamGet(pref_access_string)
        self.load_model()

    def load_model(self):
        """Load the data from the preferences entry"""
        pref_entry: str = self.pref.GetString("CustomRepositories", "")

        # The entry is saved as a space- and newline-delimited text block: break it into its
        # constituent parts
        lines = pref_entry.split("\n")
        self.model = []
        for line in lines:
            if not line:
                continue
            split_data = line.split()
            if len(split_data) > 1:
                branch = split_data[1]
            else:
                branch = "master"
            url = split_data[0]
            self.model.append([url, branch])

    def save_model(self):
        """Save the data into a preferences entry"""
        entry = ""
        for row in self.model:
            entry += f"{row[0]} {row[1]}\n"
        self.pref.SetString("CustomRepositories", entry)

    def rowCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        """The number of rows"""
        if parent.isValid():
            return 0
        return len(self.model)

    def columnCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        """The number of columns (which is always 2)"""
        if parent.isValid():
            return 0
        return 2

    def data(self, index, role=QtCore.Qt.DisplayRole):
        """The data at an index."""
        if role != QtCore.Qt.DisplayRole:
            return None
        row = index.row()
        column = index.column()
        if row > len(self.model):
            return None
        if column > 1:
            return None
        return self.model[row][column]

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        """Get the row and column header data."""
        if role != QtCore.Qt.DisplayRole:
            return None
        if orientation == QtCore.Qt.Vertical:
            return section + 1
        if section == 0:
            return translate(
                "AddonsInstaller",
                "Repository URL",
                "Preferences header for custom repositories",
            )
        if section == 1:
            return translate(
                "AddonsInstaller",
                "Branch name",
                "Preferences header for custom repositories",
            )
        return None

    def removeRows(self, row, count, parent):
        """Remove rows"""
        self.beginRemoveRows(parent, row, row + count - 1)
        for _ in range(count):
            self.model.pop(row)
        self.endRemoveRows()

    def insertRows(self, row, count, parent):
        """Insert blank rows"""
        self.beginInsertRows(parent, row, row + count - 1)
        for _ in range(count):
            self.model.insert(["", ""])
        self.endInsertRows()

    def appendData(self, url, branch):
        """Append this url and branch to the end of the list"""
        row = self.rowCount()
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        self.model.append([url, branch])
        self.endInsertRows()

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        """Set the data at this index"""
        if role != QtCore.Qt.EditRole:
            return
        self.model[index.row()][index.column()] = value
        self.dataChanged.emit(index, index)


class CustomRepositoryDialog:
    """A dialog for setting up a custom repository, with branch information"""

    def __init__(self):
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(
                os.path.dirname(__file__), "AddonManagerOptions_AddCustomRepository.ui"
            )
        )

    def exec(self):
        """Run the dialog modally, and return either None or a tuple or (url,branch)"""
        result = self.dialog.exec()
        if result == QDialog.Accepted:
            url = self.dialog.urlLineEdit.text()
            branch = self.dialog.branchLineEdit.text()
            return (url, branch)
        return (None, None)



# ***************************************************************************
# *   Copyright (c) 2023 Werner Mayer <wmayer[at]users.sourceforge.net>     *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************
"""Unit tests for the Draft Workbench, array tests."""
## @package test_array
# \ingroup drafttests
# \brief Unit tests for the Draft Workbench, array tests.

## \addtogroup drafttests
# @{
import unittest
import math

import FreeCAD as App
import Draft

from FreeCAD import Vector
from draftutils.messages import _msg


class DraftArray(unittest.TestCase):
    """Test Draft array functions."""

    def setUp(self):
        """Set up a new document to hold the tests.

        This is executed before every test, so we create a document
        to hold the objects.
        """
        doc_name = self.__class__.__name__
        if App.ActiveDocument:
            if App.ActiveDocument.Name != doc_name:
                App.newDocument(doc_name)
        else:
            App.newDocument(doc_name)
        App.setActiveDocument(doc_name)
        self.doc = App.ActiveDocument
        _msg("  Temporary document '{}'".format(self.doc.Name))

    def test_link_array(self):
        """Create a link array."""
        box = self.doc.addObject("Part::Box","Box")
        box.Label = "Box"
        self.doc.recompute()

        array = Draft.make_ortho_array(box, v_x=App.Vector(100.0, 0.0, 0.0),
                                            v_y=App.Vector(0.0, 100.0, 0.0),
                                            v_z=App.Vector(0.0, 0.0, 100.0),
                                            n_x=12, n_y=1, n_z=1, use_link=True)

        Draft.autogroup(array)
        array.ExpandArray = True
        array.Fuse = False
        self.doc.recompute(None,True,True)

        array.NumberX = 6
        self.doc.recompute(None,True,True)
        self.assertEqual(array.Count, array.NumberX)

        array.NumberX = 24
        self.doc.recompute(None,True,True)
        self.assertEqual(array.Count, array.NumberX)

        self.doc.recompute(None,True,True)
        self.assertEqual(array.Count, array.NumberX)

    def tearDown(self):
        """Finish the test.

        This is executed after each test, so we close the document.
        """
        App.closeDocument(self.doc.Name)

## @}



# ***************************************************************************
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Initialization of the Arch workbench (graphical interface)."""

import os
import FreeCAD
import FreeCADGui


class ArchWorkbench(FreeCADGui.Workbench):
    """The Arch workbench definition."""

    def __init__(self):
        def QT_TRANSLATE_NOOP(context, text):
            return text

        __dirname__ = os.path.join(FreeCAD.getResourceDir(), "Mod", "Arch")
        _tooltip = ("The Arch workbench is used to model "
                    "architectural components, and entire buildings")
        self.__class__.Icon = os.path.join(__dirname__,
                                           "Resources", "icons",
                                           "ArchWorkbench.svg")
        self.__class__.MenuText = QT_TRANSLATE_NOOP("Arch", "Arch")
        self.__class__.ToolTip = QT_TRANSLATE_NOOP("Arch", _tooltip)

    def Initialize(self):
        """When the workbench is first loaded."""

        def QT_TRANSLATE_NOOP(context, text):
            return text

        import Draft_rc
        import DraftTools
        import DraftGui
        import Arch_rc
        import Arch

        # Load Reinforcement WB translations
        try:
            import RebarTools
            RebarTools.load_translations()
        except Exception:
            pass

        from ArchStructure import _ArchStructureGroupCommand
        from ArchAxis import _ArchAxisGroupCommand
        from ArchPanel import CommandPanelGroup
        from ArchMaterial import _ArchMaterialToolsCommand
        from ArchPipe import _ArchPipeGroupCommand

        stru_group = _ArchStructureGroupCommand
        axis_group = _ArchAxisGroupCommand
        pan_group  = CommandPanelGroup
        mat_group  = _ArchMaterialToolsCommand
        pipe_group = _ArchPipeGroupCommand

        # Set up command lists
        self.archtools = ["Arch_Wall",
                          ([QT_TRANSLATE_NOOP("Workbench", "Structure tools")],
                              list(stru_group.GetCommands(stru_group))), # tuple len=2: submenu
                          ("Arch_StructureTools", ),                     # tuple len=1: toolbar flyout
                          "Arch_Rebar_Submenu",      # will be replaced or removed
                          "Arch_Rebar",              # may be replaced
                          "Arch_CurtainWall",
                          "Arch_BuildingPart",
                          "Arch_Project",
                          "Arch_Site",
                          "Arch_Building",
                          "Arch_Floor",
                          "Arch_Reference",
                          "Arch_Window",
                          "Arch_Roof",
                          ([QT_TRANSLATE_NOOP("Workbench", "Axis tools")],
                              list(axis_group.GetCommands(axis_group))),
                          ("Arch_AxisTools", ),
                          "Arch_SectionPlane",
                          "Arch_Space",
                          "Arch_Stairs",
                          ([QT_TRANSLATE_NOOP("Workbench", "Panel tools")],
                              list(pan_group.GetCommands(pan_group))),
                          ("Arch_PanelTools", ),
                          "Arch_Equipment",
                          "Arch_Frame",
                          "Arch_Fence",
                          "Arch_Truss",
                          "Arch_Profile",
                          ([QT_TRANSLATE_NOOP("Workbench", "Material tools")],
                              list(mat_group.GetCommands(mat_group))),
                          ("Arch_MaterialTools", ),
                          "Arch_Schedule",
                          ([QT_TRANSLATE_NOOP("Workbench", "Pipe tools")],
                              list(pipe_group.GetCommands(pipe_group))),
                          ("Arch_PipeTools", ),
                          "Arch_CutPlane",
                          "Arch_CutLine",
                          "Arch_Add",
                          "Arch_Remove",
                          "Arch_Survey"]

        self.utilities = ["Arch_Component",
                          "Arch_CloneComponent",
                          "Arch_SplitMesh",
                          "Arch_MeshToShape",
                          "Arch_SelectNonSolidMeshes",
                          "Arch_RemoveShape",
                          "Arch_CloseHoles",
                          "Arch_MergeWalls",
                          "Arch_Check",
                          "Arch_ToggleIfcBrepFlag",
                          "Arch_3Views",
                          "Arch_IfcSpreadsheet",
                          "Arch_ToggleSubs"]

        # Add the rebar tools from the Reinforcement addon, if available
        try:
            import RebarTools
        except Exception:
            del self.archtools[3] # remove "Arch_Rebar_Submenu"
        else:
            class RebarGroupCommand:
                def GetCommands(self):
                    return tuple(RebarTools.RebarCommands + ["Arch_Rebar"])

                def GetResources(self):
                    return {'MenuText': QT_TRANSLATE_NOOP("Arch_RebarTools", "Rebar tools"),
                            'ToolTip': QT_TRANSLATE_NOOP("Arch_RebarTools",
                                                         "Create various types of rebars, "
                                                         "including U-shaped, L-shaped, and stirrup")}

                def IsActive(self):
                    return not FreeCAD.ActiveDocument is None
            FreeCADGui.addCommand('Arch_RebarTools', RebarGroupCommand())
            self.archtools[3] = ([QT_TRANSLATE_NOOP("Workbench", "Rebar tools")],
                                    RebarTools.RebarCommands + ["Arch_Rebar"])
            self.archtools[4] = ("Arch_RebarTools", )

        # Set up Draft command lists
        import draftutils.init_tools as it
        self.draft_drawing_commands = it.get_draft_drawing_commands()
        self.draft_annotation_commands = it.get_draft_annotation_commands()
        self.draft_modification_commands = it.get_draft_modification_commands()
        self.draft_utility_commands = it.get_draft_utility_commands_menu()
        self.draft_context_commands = it.get_draft_context_commands()
        self.draft_snap_commands = it.get_draft_snap_commands()

        # Set up toolbars
        it.init_toolbar(self,
                        QT_TRANSLATE_NOOP("Workbench", "Arch tools"),
                        self.archtools)
        it.init_toolbar(self,
                        QT_TRANSLATE_NOOP("Workbench", "Draft creation tools"),
                        self.draft_drawing_commands)
        it.init_toolbar(self,
                        QT_TRANSLATE_NOOP("Workbench", "Draft annotation tools"),
                        self.draft_annotation_commands)
        it.init_toolbar(self,
                        QT_TRANSLATE_NOOP("Workbench", "Draft modification tools"),
                        self.draft_modification_commands)
        it.init_toolbar(self,
                        QT_TRANSLATE_NOOP("Workbench", "Draft snap"),
                        self.draft_snap_commands)

        # Set up menus
        it.init_menu(self,
                     [QT_TRANSLATE_NOOP("Workbench", "&Arch"),
                      QT_TRANSLATE_NOOP("Workbench", "Utilities")],
                     self.utilities)
        it.init_menu(self,
                     [QT_TRANSLATE_NOOP("Workbench", "&Arch")],
                     self.archtools)
        it.init_menu(self,
                     [QT_TRANSLATE_NOOP("Workbench", "&Draft"),
                      QT_TRANSLATE_NOOP("Workbench", "Creation")],
                     self.draft_drawing_commands)
        it.init_menu(self,
                     [QT_TRANSLATE_NOOP("Workbench", "&Draft"),
                      QT_TRANSLATE_NOOP("Workbench", "Annotation")],
                     self.draft_annotation_commands)
        it.init_menu(self,
                     [QT_TRANSLATE_NOOP("Workbench", "&Draft"),
                      QT_TRANSLATE_NOOP("Workbench", "Modification")],
                     self.draft_modification_commands)
        it.init_menu(self,
                     [QT_TRANSLATE_NOOP("Workbench", "&Draft"),
                      QT_TRANSLATE_NOOP("Workbench", "Utilities")],
                     self.draft_utility_commands)

        FreeCADGui.addIconPath(":/icons")
        FreeCADGui.addLanguagePath(":/translations")

        # Set up preferences pages
        if hasattr(FreeCADGui, "draftToolBar"):
            if not hasattr(FreeCADGui.draftToolBar, "loadedArchPreferences"):
                FreeCADGui.addPreferencePage(":/ui/preferences-arch.ui", QT_TRANSLATE_NOOP("QObject", "Arch"))
                FreeCADGui.addPreferencePage(":/ui/preferences-archdefaults.ui", QT_TRANSLATE_NOOP("QObject", "Arch"))
                FreeCADGui.draftToolBar.loadedArchPreferences = True
            if not hasattr(FreeCADGui.draftToolBar, "loadedPreferences"):
                FreeCADGui.addPreferencePage(":/ui/preferences-draft.ui", QT_TRANSLATE_NOOP("QObject", "Draft"))
                FreeCADGui.addPreferencePage(":/ui/preferences-draftsnap.ui", QT_TRANSLATE_NOOP("QObject", "Draft"))
                FreeCADGui.addPreferencePage(":/ui/preferences-draftvisual.ui", QT_TRANSLATE_NOOP("QObject", "Draft"))
                FreeCADGui.addPreferencePage(":/ui/preferences-drafttexts.ui", QT_TRANSLATE_NOOP("QObject", "Draft"))
                FreeCADGui.draftToolBar.loadedPreferences = True

        FreeCAD.Console.PrintLog('Loading Arch workbench, done.\n')

    def Activated(self):
        """When entering the workbench."""
        if hasattr(FreeCADGui, "draftToolBar"):
            FreeCADGui.draftToolBar.Activated()
        if hasattr(FreeCADGui, "Snapper"):
            FreeCADGui.Snapper.show()
        FreeCAD.Console.PrintLog("Arch workbench activated.\n")

    def Deactivated(self):
        """When leaving the workbench."""
        if hasattr(FreeCADGui, "draftToolBar"):
            FreeCADGui.draftToolBar.Deactivated()
        if hasattr(FreeCADGui, "Snapper"):
            FreeCADGui.Snapper.hide()
        FreeCAD.Console.PrintLog("Arch workbench deactivated.\n")

    def ContextMenu(self, recipient):
        """Define an optional custom context menu."""
        self.appendContextMenu("Utilities", self.draft_context_commands)

    def GetClassName(self):
        """Type of workbench."""
        return "Gui::PythonWorkbench"


FreeCADGui.addWorkbench(ArchWorkbench)

# Preference pages for importing and exporting various file formats
# are independent of the loading of the workbench and can be loaded at startup
import Arch_rc
from PySide.QtCore import QT_TRANSLATE_NOOP
FreeCADGui.addPreferencePage(":/ui/preferences-ifc.ui", QT_TRANSLATE_NOOP("QObject", "Import-Export"))
FreeCADGui.addPreferencePage(":/ui/preferences-ifc-export.ui", QT_TRANSLATE_NOOP("QObject", "Import-Export"))
FreeCADGui.addPreferencePage(":/ui/preferences-dae.ui", QT_TRANSLATE_NOOP("QObject", "Import-Export"))

FreeCAD.__unit_test__ += ["TestArch"]



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Classes to manage "Developer Mode" """

import os
import datetime
import subprocess

import FreeCAD
import FreeCADGui

from PySide.QtWidgets import (
    QFileDialog,
    QListWidgetItem,
    QDialog,
    QSizePolicy,
    QMessageBox,
)
from PySide.QtGui import (
    QIcon,
    QPixmap,
)
from PySide.QtCore import Qt
from addonmanager_git import GitManager, NoGitFound

from addonmanager_devmode_add_content import AddContent
from addonmanager_devmode_validators import NameValidator, VersionValidator
from addonmanager_devmode_predictor import Predictor
from addonmanager_devmode_people_table import PeopleTable
from addonmanager_devmode_licenses_table import LicensesTable
import addonmanager_utilities as utils

translate = FreeCAD.Qt.translate

# pylint: disable=too-few-public-methods

ContentTypeRole = Qt.UserRole
ContentIndexRole = Qt.UserRole + 1


class AddonGitInterface:
    """Wrapper to handle the git calls needed by this class"""

    git_manager = None

    def __init__(self, path):
        self.git_exists = False
        if not AddonGitInterface.git_manager:
            try:
                AddonGitInterface.git_manager = GitManager()
            except NoGitFound:
                FreeCAD.Console.PrintLog(
                    "No git found, Addon Manager Developer Mode disabled."
                )
                return

        self.path = path
        if os.path.exists(os.path.join(path, ".git")):
            self.git_exists = True
            self.branch = AddonGitInterface.git_manager.current_branch(self.path)
            self.remote = AddonGitInterface.git_manager.get_remote(self.path)

    @property
    def branches(self):
        """The branches available for this repo."""
        if self.git_exists:
            return AddonGitInterface.git_manager.get_branches(self.path)
        return []

    @property
    def committers(self):
        """The committers to this repo, in the last ten commits"""
        if self.git_exists:
            return AddonGitInterface.git_manager.get_last_committers(self.path, 10)
        return []

    @property
    def authors(self):
        """The committers to this repo, in the last ten commits"""
        if self.git_exists:
            return AddonGitInterface.git_manager.get_last_authors(self.path, 10)
        return []


# pylint: disable=too-many-instance-attributes


class DeveloperMode:
    """The main Developer Mode dialog, for editing package.xml metadata graphically."""

    def __init__(self):
        # In the UI we want to show a translated string for the person type, but the underlying
        # string must be the one expected by the metadata parser, in English
        self.person_type_translation = {
            "maintainer": translate("AddonsInstaller", "Maintainer"),
            "author": translate("AddonsInstaller", "Author"),
        }
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "developer_mode.ui")
        )
        self.people_table = PeopleTable()
        self.licenses_table = LicensesTable()
        large_size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        large_size_policy.setHorizontalStretch(2)
        small_size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        small_size_policy.setHorizontalStretch(1)
        self.people_table.widget.setSizePolicy(large_size_policy)
        self.licenses_table.widget.setSizePolicy(small_size_policy)
        self.dialog.peopleAndLicenseshorizontalLayout.addWidget(
            self.people_table.widget
        )
        self.dialog.peopleAndLicenseshorizontalLayout.addWidget(
            self.licenses_table.widget
        )
        self.pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        self.current_mod: str = ""
        self.git_interface = None
        self.has_toplevel_icon = False
        self.metadata = None

        self._setup_dialog_signals()

        self.dialog.displayNameLineEdit.setValidator(NameValidator())
        self.dialog.versionLineEdit.setValidator(VersionValidator())
        self.dialog.minPythonLineEdit.setValidator(VersionValidator())

        self.dialog.addContentItemToolButton.setIcon(
            QIcon.fromTheme("add", QIcon(":/icons/list-add.svg"))
        )
        self.dialog.removeContentItemToolButton.setIcon(
            QIcon.fromTheme("remove", QIcon(":/icons/list-remove.svg"))
        )

    def show(self, parent=None, path: str = None):
        """Show the main dev mode dialog"""
        if parent:
            self.dialog.setParent(parent)
        if path and os.path.exists(path):
            self.dialog.pathToAddonComboBox.setEditText(path)
        elif self.pref.HasGroup("recentModsList"):
            recent_mods_group = self.pref.GetGroup("recentModsList")
            entry = recent_mods_group.GetString("Mod0", "")
            if entry:
                self._populate_dialog(entry)
                self._update_recent_mods(entry)
                self._populate_combo()
            else:
                self._clear_all_fields()
        else:
            self._clear_all_fields()

        result = self.dialog.exec()
        if result == QDialog.Accepted:
            self._sync_metadata_to_ui()
            now = datetime.date.today()
            self.metadata.Date = str(now)
            self.metadata.write(os.path.join(self.current_mod, "package.xml"))

    def _populate_dialog(self, path_to_repo: str):
        """Populate this dialog using the best available parsing of the contents of the repo at
        path_to_repo. This is a multi-layered process that starts with any existing package.xml
        file or other known metadata files, and proceeds through examining the contents of the
        directory structure."""
        self.current_mod = path_to_repo
        self._scan_for_git_info(self.current_mod)

        metadata_path = os.path.join(path_to_repo, "package.xml")
        if os.path.exists(metadata_path):
            try:
                self.metadata = FreeCAD.Metadata(metadata_path)
            except FreeCAD.Base.XMLBaseException as e:
                FreeCAD.Console.PrintError(
                    translate(
                        "AddonsInstaller",
                        "XML failure while reading metadata from file {}",
                    ).format(metadata_path)
                    + "\n\n"
                    + str(e)
                    + "\n\n"
                )
            except FreeCAD.Base.RuntimeError as e:
                FreeCAD.Console.PrintError(
                    translate("AddonsInstaller", "Invalid metadata in file {}").format(
                        metadata_path
                    )
                    + "\n\n"
                    + str(e)
                    + "\n\n"
                )

        self._clear_all_fields()

        if not self.metadata:
            self._predict_metadata()

        self.dialog.displayNameLineEdit.setText(self.metadata.Name)
        self.dialog.descriptionTextEdit.setPlainText(self.metadata.Description)
        self.dialog.versionLineEdit.setText(self.metadata.Version)
        self.dialog.minPythonLineEdit.setText(self.metadata.PythonMin)

        self._populate_urls_from_metadata(self.metadata)
        self._populate_contents_from_metadata(self.metadata)

        self._populate_icon_from_metadata(self.metadata)

        self.people_table.show(self.metadata)
        self.licenses_table.show(self.metadata, self.current_mod)

    def _populate_urls_from_metadata(self, metadata):
        """Use the passed metadata object to populate the urls"""
        for url in metadata.Urls:
            if url["type"] == "website":
                self.dialog.websiteURLLineEdit.setText(url["location"])
            elif url["type"] == "repository":
                self.dialog.repositoryURLLineEdit.setText(url["location"])
                branch_from_metadata = url["branch"]
                branch_from_local_path = self.git_interface.branch
                if branch_from_metadata != branch_from_local_path:
                    # pylint: disable=line-too-long
                    FreeCAD.Console.PrintWarning(
                        translate(
                            "AddonsInstaller",
                            "WARNING: Path specified in package.xml metadata does not match currently checked-out branch.",
                        )
                        + "\n"
                    )
                self.dialog.branchComboBox.setCurrentText(branch_from_metadata)
            elif url["type"] == "bugtracker":
                self.dialog.bugtrackerURLLineEdit.setText(url["location"])
            elif url["type"] == "readme":
                self.dialog.readmeURLLineEdit.setText(url["location"])
            elif url["type"] == "documentation":
                self.dialog.documentationURLLineEdit.setText(url["location"])
            elif url["type"] == "discussion":
                self.dialog.discussionURLLineEdit.setText(url["location"])

    def _populate_contents_from_metadata(self, metadata):
        """Use the passed metadata object to populate the contents list"""
        contents = metadata.Content
        self.dialog.contentsListWidget.clear()
        for content_type in contents:
            counter = 0
            for item in contents[content_type]:
                contents_string = f"[{content_type}] "
                info = []
                if item.Name:
                    info.append(translate("AddonsInstaller", "Name") + ": " + item.Name)
                if item.Classname:
                    info.append(
                        translate("AddonsInstaller", "Class") + ": " + item.Classname
                    )
                if item.Description:
                    info.append(
                        translate("AddonsInstaller", "Description")
                        + ": "
                        + item.Description
                    )
                if item.Subdirectory:
                    info.append(
                        translate("AddonsInstaller", "Subdirectory")
                        + ": "
                        + item.Subdirectory
                    )
                if item.File:
                    info.append(
                        translate("AddonsInstaller", "Files")
                        + ": "
                        + ", ".join(item.File)
                    )
                contents_string += ", ".join(info)

                item = QListWidgetItem(contents_string)
                item.setData(ContentTypeRole, content_type)
                item.setData(ContentIndexRole, counter)
                self.dialog.contentsListWidget.addItem(item)
                counter += 1

    def _populate_icon_from_metadata(self, metadata):
        """Use the passed metadata object to populate the icon fields"""
        self.dialog.iconDisplayLabel.setPixmap(QPixmap())
        icon = metadata.Icon
        icon_path = None
        if icon:
            icon_path = os.path.join(self.current_mod, icon.replace("/", os.path.sep))
            self.has_toplevel_icon = True
        else:
            self.has_toplevel_icon = False
            contents = metadata.Content
            if "workbench" in contents:
                for wb in contents["workbench"]:
                    icon = wb.Icon
                    path = wb.Subdirectory
                    if icon:
                        icon_path = os.path.join(
                            self.current_mod, path, icon.replace("/", os.path.sep)
                        )
                        break

        if icon_path and os.path.isfile(icon_path):
            icon_data = QIcon(icon_path)
            if not icon_data.isNull():
                self.dialog.iconDisplayLabel.setPixmap(icon_data.pixmap(32, 32))
        self.dialog.iconPathLineEdit.setText(icon)

    def _predict_metadata(self):
        """If there is no metadata, try to guess at values for it"""
        self.metadata = FreeCAD.Metadata()
        predictor = Predictor()
        self.metadata = predictor.predict_metadata(self.current_mod)

    def _scan_for_git_info(self, path: str):
        """Look for branch availability"""
        self.git_interface = AddonGitInterface(path)
        if self.git_interface.git_exists:
            self.dialog.branchComboBox.clear()
            for branch in self.git_interface.branches:
                if branch and branch.startswith("origin/") and branch != "origin/HEAD":
                    self.dialog.branchComboBox.addItem(branch[len("origin/") :])
            self.dialog.branchComboBox.setCurrentText(self.git_interface.branch)

    def _clear_all_fields(self):
        """Clear out all fields"""
        self.dialog.displayNameLineEdit.clear()
        self.dialog.descriptionTextEdit.clear()
        self.dialog.versionLineEdit.clear()
        self.dialog.websiteURLLineEdit.clear()
        self.dialog.repositoryURLLineEdit.clear()
        self.dialog.bugtrackerURLLineEdit.clear()
        self.dialog.readmeURLLineEdit.clear()
        self.dialog.documentationURLLineEdit.clear()
        self.dialog.discussionURLLineEdit.clear()
        self.dialog.minPythonLineEdit.clear()
        self.dialog.iconDisplayLabel.setPixmap(QPixmap())
        self.dialog.iconPathLineEdit.clear()

    def _setup_dialog_signals(self):
        """Set up the signal and slot connections for the main dialog."""

        self.dialog.addonPathBrowseButton.clicked.connect(
            self._addon_browse_button_clicked
        )
        self.dialog.pathToAddonComboBox.editTextChanged.connect(
            self._addon_combo_text_changed
        )
        self.dialog.detectMinPythonButton.clicked.connect(
            self._detect_min_python_clicked
        )
        self.dialog.iconBrowseButton.clicked.connect(self._browse_for_icon_clicked)

        self.dialog.addContentItemToolButton.clicked.connect(self._add_content_clicked)
        self.dialog.removeContentItemToolButton.clicked.connect(
            self._remove_content_clicked
        )
        self.dialog.contentsListWidget.itemSelectionChanged.connect(
            self._content_selection_changed
        )
        self.dialog.contentsListWidget.itemDoubleClicked.connect(self._edit_content)

        self.dialog.versionToTodayButton.clicked.connect(self._set_to_today_clicked)

        # Finally, populate the combo boxes, etc.
        self._populate_combo()

        # Disable all the "Remove" buttons until something is selected
        self.dialog.removeContentItemToolButton.setDisabled(True)

    def _sync_metadata_to_ui(self):
        """Take the data from the UI fields and put it into the stored metadata
        object. Only overwrites known data fields: unknown metadata will be retained."""

        if not self.metadata:
            self.metadata = FreeCAD.Metadata()

        self.metadata.Name = self.dialog.displayNameLineEdit.text()
        self.metadata.Description = (
            self.dialog.descriptionTextEdit.document().toPlainText()
        )
        self.metadata.Version = self.dialog.versionLineEdit.text()
        self.metadata.Icon = self.dialog.iconPathLineEdit.text()

        urls = []
        if self.dialog.websiteURLLineEdit.text():
            urls.append(
                {"location": self.dialog.websiteURLLineEdit.text(), "type": "website"}
            )
        if self.dialog.repositoryURLLineEdit.text():
            urls.append(
                {
                    "location": self.dialog.repositoryURLLineEdit.text(),
                    "type": "repository",
                    "branch": self.dialog.branchComboBox.currentText(),
                }
            )
        if self.dialog.bugtrackerURLLineEdit.text():
            urls.append(
                {
                    "location": self.dialog.bugtrackerURLLineEdit.text(),
                    "type": "bugtracker",
                }
            )
        if self.dialog.readmeURLLineEdit.text():
            urls.append(
                {"location": self.dialog.readmeURLLineEdit.text(), "type": "readme"}
            )
        if self.dialog.documentationURLLineEdit.text():
            urls.append(
                {
                    "location": self.dialog.documentationURLLineEdit.text(),
                    "type": "documentation",
                }
            )
        if self.dialog.discussionURLLineEdit.text():
            urls.append(
                {
                    "location": self.dialog.discussionURLLineEdit.text(),
                    "type": "discussion",
                }
            )
        self.metadata.Urls = urls

        if self.dialog.minPythonLineEdit.text():
            self.metadata.PythonMin = self.dialog.minPythonLineEdit.text()
        else:
            self.metadata.PythonMin = "0.0.0"  # Code for "unset"

        # Content, people, and licenses should already be synchronized

    ###############################################################################################
    #                                         DIALOG SLOTS
    ###############################################################################################

    def _addon_browse_button_clicked(self):
        """Launch a modal file/folder selection dialog -- if something is selected, it is
        processed by the parsing code and used to fill in the contents of the rest of the
        dialog."""

        start_dir = os.path.join(FreeCAD.getUserAppDataDir(), "Mod")
        mod_dir = QFileDialog.getExistingDirectory(
            parent=self.dialog,
            caption=translate(
                "AddonsInstaller",
                "Select the folder containing your Addon",
            ),
            dir=start_dir,
        )

        if mod_dir and os.path.exists(mod_dir):
            self.dialog.pathToAddonComboBox.setEditText(mod_dir)

    def _addon_combo_text_changed(self, new_text: str):
        """Called when the text is changed, either because it was directly edited, or because
        a new item was selected."""
        if new_text == self.current_mod:
            # It doesn't look like it actually changed, bail out
            return
        self.metadata = None
        self._clear_all_fields()
        if not os.path.exists(new_text):
            # This isn't a thing (Yet. Maybe the user is still typing?)
            return
        self._populate_dialog(new_text)
        self._update_recent_mods(new_text)
        self._populate_combo()

    def _populate_combo(self):
        """Fill in the combo box with the values from the stored recent mods list, selecting the
        top one. Does not trigger any signals."""
        combo = self.dialog.pathToAddonComboBox
        combo.blockSignals(True)
        recent_mods_group = self.pref.GetGroup("recentModsList")
        recent_mods = set()
        combo.clear()
        for i in range(10):
            entry_name = f"Mod{i}"
            mod = recent_mods_group.GetString(entry_name, "None")
            if mod != "None" and mod not in recent_mods and os.path.exists(mod):
                recent_mods.add(mod)
                combo.addItem(mod)
        if recent_mods:
            combo.setCurrentIndex(0)
        combo.blockSignals(False)

    def _update_recent_mods(self, path):
        """Update the list of recent mods, storing at most ten, with path at the top of the
        list."""
        recent_mod_paths = [path]
        if self.pref.HasGroup("recentModsList"):
            recent_mods_group = self.pref.GetGroup("recentModsList")

            # This group has a maximum of ten entries, sorted by last-accessed date
            for i in range(0, 10):
                entry_name = f"Mod{i}"
                entry = recent_mods_group.GetString(entry_name, "")
                if entry and entry not in recent_mod_paths and os.path.exists(entry):
                    recent_mod_paths.append(entry)

            # Remove the whole thing, so we can recreate it from scratch
            self.pref.RemGroup("recentModsList")

        if recent_mod_paths:
            recent_mods_group = self.pref.GetGroup("recentModsList")
            for i, mod in zip(range(10), recent_mod_paths):
                entry_name = f"Mod{i}"
                recent_mods_group.SetString(entry_name, mod)

    def _add_content_clicked(self):
        """Callback: The Add Content button was clicked"""
        dlg = AddContent(self.current_mod, self.metadata)
        singleton = False
        if self.dialog.contentsListWidget.count() == 0:
            singleton = True
        content_type, new_metadata = dlg.exec(singleton=singleton)
        if content_type and new_metadata:
            self.metadata.addContentItem(content_type, new_metadata)
            self._populate_contents_from_metadata(self.metadata)

    def _remove_content_clicked(self):
        """Callback: the remove content button was clicked"""

        item = self.dialog.contentsListWidget.currentItem()
        if not item:
            return
        content_type = item.data(ContentTypeRole)
        content_index = item.data(ContentIndexRole)
        if self.metadata.Content[content_type] and content_index < len(
            self.metadata.Content[content_type]
        ):
            content_name = self.metadata.Content[content_type][content_index].Name
            self.metadata.removeContentItem(content_type, content_name)
            self._populate_contents_from_metadata(self.metadata)

    def _content_selection_changed(self):
        """Callback: the selected content item changed"""
        items = self.dialog.contentsListWidget.selectedItems()
        if items:
            self.dialog.removeContentItemToolButton.setDisabled(False)
        else:
            self.dialog.removeContentItemToolButton.setDisabled(True)

    def _edit_content(self, item):
        """Callback: a content row was double-clicked"""
        dlg = AddContent(self.current_mod, self.metadata)

        content_type = item.data(ContentTypeRole)
        content_index = item.data(ContentIndexRole)

        content = self.metadata.Content
        metadata = content[content_type][content_index]
        old_name = metadata.Name
        new_type, new_metadata = dlg.exec(content_type, metadata, len(content) == 1)
        if new_type and new_metadata:
            self.metadata.removeContentItem(content_type, old_name)
            self.metadata.addContentItem(new_type, new_metadata)
            self._populate_contents_from_metadata(self.metadata)

    def _set_to_today_clicked(self):
        """Callback: the "set to today" button was clicked"""
        year = datetime.date.today().year
        month = datetime.date.today().month
        day = datetime.date.today().day
        version_string = f"{year}.{month:>02}.{day:>02}"
        self.dialog.versionLineEdit.setText(version_string)

    def _detect_min_python_clicked(self):
        if not self._ensure_vermin_loaded():
            FreeCAD.Console.PrintWarning(
                translate(
                    "AddonsInstaller",
                    "No Vermin, cancelling operation.",
                    "NOTE: Vermin is a Python package and proper noun - do not translate",
                )
                + "\n"
            )
            return
        FreeCAD.Console.PrintMessage(
            translate(
                "AddonsInstaller", "Scanning Addon for Python version compatibility"
            )
            + "...\n"
        )
        # pylint: disable=import-outside-toplevel
        import vermin

        required_minor_version = 0
        for dir_path, _, filenames in os.walk(self.current_mod):
            for filename in filenames:
                if filename.endswith(".py"):
                    with open(os.path.join(dir_path, filename), encoding="utf-8") as f:
                        contents = f.read()
                        version_strings = vermin.version_strings(
                            vermin.detect(contents)
                        )
                        version = version_strings.split(",")
                        if len(version) >= 2:
                            # Only care about Py3, and only if there is a dot in the version:
                            if "." in version[1]:
                                py3 = version[1].split(".")
                                major = int(py3[0].strip())
                                minor = int(py3[1].strip())
                                if major == 3:
                                    FreeCAD.Console.PrintLog(
                                        f"Detected Python 3.{minor} required by {filename}\n"
                                    )
                                    required_minor_version = max(
                                        required_minor_version, minor
                                    )
        self.dialog.minPythonLineEdit.setText(f"3.{required_minor_version}")
        QMessageBox.information(
            self.dialog,
            translate("AddonsInstaller", "Minimum Python Version Detected"),
            translate(
                "AddonsInstaller",
                "Vermin auto-detected a required version of Python 3.{}",
            ).format(required_minor_version),
            QMessageBox.Ok,
        )

    def _ensure_vermin_loaded(self) -> bool:
        try:
            # pylint: disable=import-outside-toplevel,unused-import
            import vermin
        except ImportError:
            # pylint: disable=line-too-long
            response = QMessageBox.question(
                self.dialog,
                translate("AddonsInstaller", "Install Vermin?"),
                translate(
                    "AddonsInstaller",
                    "Auto-detecting the required version of Python for this Addon requires Vermin (https://pypi.org/project/vermin/). OK to install?",
                ),
                QMessageBox.Yes | QMessageBox.Cancel,
            )
            if response == QMessageBox.Cancel:
                return False
            FreeCAD.Console.PrintMessage(
                translate("AddonsInstaller", "Attempting to install Vermin from PyPi")
                + "...\n"
            )
            python_exe = utils.get_python_exe()
            vendor_path = os.path.join(
                FreeCAD.getUserAppDataDir(), "AdditionalPythonPackages"
            )
            if not os.path.exists(vendor_path):
                os.makedirs(vendor_path)

            proc = subprocess.run(
                [
                    python_exe,
                    "-m",
                    "pip",
                    "install",
                    "--disable-pip-version-check",
                    "--target",
                    vendor_path,
                    "vermin",
                ],
                capture_output=True,
                check=True,
            )
            FreeCAD.Console.PrintMessage(proc.stdout.decode())
            if proc.returncode != 0:
                QMessageBox.critical(
                    self.dialog,
                    translate("AddonsInstaller", "Installation failed"),
                    translate(
                        "AddonsInstaller",
                        "Failed to install Vermin -- check Report View for details.",
                        "'Vermin' is the name of a Python package, do not translate",
                    ),
                    QMessageBox.Cancel,
                )
                return False
        try:
            # pylint: disable=import-outside-toplevel
            import vermin
        except ImportError:
            QMessageBox.critical(
                self.dialog,
                translate("AddonsInstaller", "Installation failed"),
                translate(
                    "AddonsInstaller",
                    "Failed to import vermin after installation -- cannot scan Addon.",
                    "'vermin' is the name of a Python package, do not translate",
                ),
                QMessageBox.Cancel,
            )
            return False
        return True

    def _browse_for_icon_clicked(self):
        """Callback: when the "Browse..." button for the icon field is clicked"""
        new_icon_path, _ = QFileDialog.getOpenFileName(
            parent=self.dialog,
            caption=translate(
                "AddonsInstaller",
                "Select an icon file for this package",
            ),
            dir=self.current_mod,
        )

        if not new_icon_path:
            return

        base_path = self.current_mod.replace("/", os.path.sep)
        icon_path = new_icon_path.replace("/", os.path.sep)
        if base_path[-1] != os.path.sep:
            base_path += os.path.sep

        if not icon_path.startswith(base_path):
            FreeCAD.Console.PrintError(
                translate("AddonsInstaller", "{} is not a subdirectory of {}").format(
                    new_icon_path, self.current_mod
                )
                + "\n"
            )
            return
        self.metadata.Icon = new_icon_path[len(base_path) :]
        self._populate_icon_from_metadata(self.metadata)



# ***************************************************************************
# *   Copyright (c) 2023 edi <edi271@a1.net>                                *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the TechDraw HoleShaftFit GuiCommand."""

__title__ = "TechDrawTools.CommandHoleShaftFit"
__author__ = "edi"
__url__ = "https://www.freecad.org"
__version__ = "00.01"
__date__ = "2023/02/07"

from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui

import TechDrawTools

class CommandHoleShaftFit:
    """Adds a hole or shaft fit to a selected dimension."""

    def GetResources(self):
        """Return a dictionary with data that will be used by the button or menu item."""
        return {'Pixmap': 'actions/TechDraw_HoleShaftFit.svg',
                'Accel': "",
                'MenuText': QT_TRANSLATE_NOOP("TechDraw_HoleShaftFit", "Add hole or shaft fit"),
                'ToolTip': QT_TRANSLATE_NOOP("TechDraw_HoleShaftFit", "Add a hole or shaft fit to a dimension<br>\
                - select one length dimension or diameter dimension<br>\
                - click the tool button, a panel opens<br>\
                - select shaft fit / hole fit<br>\
                - select the desired ISO 286 fit field using the combo box")}

    def Activated(self):
        """Run the following code when the command is activated (button press)."""
        sel = Gui.Selection.getSelectionEx()
        #if sel and sel[0].Object.TypeId == 'TechDraw::DrawViewDimension':
        if sel[0].Object.TypeId == 'TechDraw::DrawViewDimension':
            self.ui = TechDrawTools.TaskHoleShaftFit(sel)
            Gui.Control.showDialog(self.ui)

    def IsActive(self):
        """Return True when the command should be active or False when it should be disabled (greyed)."""
        if App.ActiveDocument:
            return TechDrawTools.TDToolsUtil.havePage() and TechDrawTools.TDToolsUtil.haveView()
        else:
            return False

#
# The command must be "registered" with a unique name by calling its class.
Gui.addCommand('TechDraw_HoleShaftFit', CommandHoleShaftFit())